#!/usr/bin/env python3
import argparse
import os
import os.path
import re
import subprocess
import sys


def run(*args, check=True):
    return subprocess.run(args, capture_output=True, check=check, universal_newlines=True)


def get_submodules():
    """
    Return list of submodules for current repo, sorted by name.

    Each item looks like:
    {
        'name': 'aws-c-common',
        'path': 'crt/aws-c-common',
        'url': 'https://github.com/awslabs/aws-c-common.git',
    }
    """
    if not os.path.exists('.gitmodules'):
        sys.exit(f'No .gitmodules found in {os.getcwd()}')

    submodules = []
    start_pattern = re.compile(r'\[submodule')
    path_pattern = re.compile(r'\s+path = (\S+)')
    url_pattern = re.compile(r'\s+url = (\S+)')

    current = None
    with open('.gitmodules', 'r') as f:
        for line in f.readlines():
            m = start_pattern.match(line)
            if m:
                current = {}
                submodules.append(current)
                continue

            m = path_pattern.match(line)
            if m:
                current['path'] = m.group(1)
                current['name'] = os.path.basename(current['path'])
                continue

            m = url_pattern.match(line)
            if m:
                current['url'] = m.group(1)
                continue

    return sorted(submodules, key=lambda x: x['name'])


def get_release_tags(prefix='v'):
    """
    Return list of release tags for current repo, sorted high to low.

    Each item looks like:
    {
        'commit': 'e18f041a0c8d17189f2eae2a32f16e0a7a3f0f1c',
        'version': 'v0.5.18'
        'num_tuple': (0,5,18),
    }
    """
    git_output = run('git', 'ls-remote', '--tags').stdout
    tags = []
    for line in git_output.splitlines():
        # line looks like: "e18f041a0c8d17189f2eae2a32f16e0a7a3f0f1c refs/tags/v0.5.18"
        match = re.match(
            r'([a-f0-9]+)\s+refs/tags/(' + prefix + r'([0-9]+)\.([0-9]+)\.([0-9]+))$', line)
        if not match:
            # skip malformed release tags
            continue
        tags.append({
            'commit': match.group(1),
            'version': match.group(2),
            'num_tuple': (int(match.group(3)), int(match.group(4)), int(match.group(5))),
        })

    # sort highest version first
    return sorted(tags, reverse=True, key=lambda tag: tag['num_tuple'])


def get_current_commit():
    git_output = run('git', 'rev-parse', 'HEAD').stdout
    return git_output.splitlines()[0]


def is_ancestor(ancestor, descendant):
    """Return whether first commit is an ancestor to the second'"""
    result = run('git', 'merge-base', '--is-ancestor',
                 ancestor, descendant, check=False)
    return result.returncode == 0


def get_tag_for_commit(tags, commit):
    for tag in tags:
        if tag['commit'] == commit:
            return tag
    return None


def main():
    parser = argparse.ArgumentParser(
        description="Update submodules to latest tags")
    parser.add_argument('ignore', nargs='*', help="submodules to ignore")
    parser.add_argument('--dry-run', action='store_true',
                        help="print without actually updating")
    args = parser.parse_args()

    root_path = os.getcwd()
    submodules = get_submodules()
    name_pad = max([len(x['name']) for x in submodules])
    for submodule in submodules:
        name = submodule['name']

        os.chdir(os.path.join(root_path, submodule['path']))

        version_prefix = 'v'
        # aws-crt-java uses FIPS releases of aws-lc
        if name == 'aws-lc' and 'aws-crt-java' in root_path:
            version_prefix = 'AWS-LC-FIPS-'

        tags = get_release_tags(version_prefix)
        current_commit = get_current_commit()
        current_tag = get_tag_for_commit(tags, current_commit)
        sync_from = current_tag['version'] if current_tag else current_commit

        if name in args.ignore:
            print(f"{name:<{name_pad}} {sync_from} (ignored)")
            continue

        latest_tag = tags[0]
        sync_to = latest_tag['version']

        # The only time we don't want to sync to the latest release is:
        # The submodule is at some commit beyond the latest release,
        # and the CRT team doesn't control this repo so can't just cut a new release
        if sync_from != sync_to and current_tag is None:
            if name in ['aws-lc', 's2n', 's2n-tls']:
                # must fetch tags before we can check their ancestry
                run('git', 'fetch', '--tags', '--prune', '--prune-tags', '--force')
                if not is_ancestor(ancestor=current_commit, descendant=sync_to):
                    sync_to = sync_from

        if sync_from == sync_to:
            print(f"{name:<{name_pad}} {sync_from} ✓")
        else:
            print(f"{name:<{name_pad}} {sync_from} -> {sync_to}")
            if not args.dry_run:
                run('git', 'fetch', '--tags', '--prune', '--prune-tags', '--force')
                run('git', 'checkout', sync_to)
                run('git', 'submodule', 'update')


if __name__ == '__main__':
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import re
import subprocess
import threading
import itertools
import random
import string
import pytest

from constants import TEST_CERT_DIRECTORY
from global_flags import get_flag, S2N_PROVIDER_VERSION


def data_bytes(n_bytes):
    """
    Generate bytes to send over the TLS connection.
    These bytes purposefully fall outside of the ascii range
    to prevent triggering "connected commands" present in
    some SSL clients.
    """
    byte_array = [0] * n_bytes
    allowed = [i for i in range(128, 255)]

    j = 0
    for i in range(n_bytes):
        byte_array[i] = allowed[j]
        j += 1
        if j > 126:
            j = 0

    return bytes(byte_array)


def random_str(n):
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(n)
    )


def pq_enabled():
    """
    Returns true or false to indicate whether PQ crypto is enabled in s2n
    """
    return "awslc" in get_flag(S2N_PROVIDER_VERSION)


class AvailablePorts(object):
    """
    This iterator will atomically return the next number.
    This is useful when running multiple tests in parallel
    that all need unique port numbers.
    """

    def __init__(self, low=8000, high=30000):
        worker_count = 1
        # If pytest is being run in parallel, worker processes will have
        # the WORKER_COUNT variable set.
        parallel_workers = os.getenv("PYTEST_XDIST_WORKER_COUNT")
        if parallel_workers is not None:
            worker_count = int(parallel_workers)
        chunk_size = int((high - low) / worker_count)

        # If xdist is being used, parse the workerid from the envvar. This can
        # be used to allocate unique ports to each worker.
        worker = os.getenv("PYTEST_XDIST_WORKER")
        worker_id = 0
        if worker is not None:
            worker_id = re.findall(r"gw(\d+)", worker)
            if len(worker_id) != 0:
                worker_id = int(worker_id[0])

        # This is a naive way to allocate ports, but it allows us to cut
        # the run time in half without workers colliding.
        worker_offset = worker_id * chunk_size
        base_range = range(low + worker_offset, high)
        wrap_range = range(low, low + worker_offset)
        self.ports = iter(itertools.chain(base_range, wrap_range))

        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.ports)


class TimeoutException(subprocess.SubprocessError):
    """
    TimeoutException wraps the subprocess class giving more control
    over the formatting of output.
    """

    def __init__(self, timeout_exception):
        self.exception = timeout_exception

    def __str__(self):
        cmd = " ".join(self.exception.cmd)
        return "{} {}".format(self.exception, cmd)


class Cert(object):
    def __init__(self, name, prefix, location=TEST_CERT_DIRECTORY):
        self.name = name
        self.cert = location + prefix + "_cert.pem"
        self.key = location + prefix + "_key.pem"
        self.algorithm = "ANY"
        self.curve = None

        if "ECDSA" in name:
            self.algorithm = "EC"
            self.curve = name[-3:]
        elif "RSA" in name:
            self.algorithm = "RSA"
        if "PSS" in name:
            self.algorithm = "RSAPSS"

    def compatible_with_cipher(self, cipher):
        if self.algorithm == cipher.algorithm:
            return True
        # TLS1.3 cipher suites do not specify auth method, so allow any auth method
        if cipher.algorithm == "ANY":
            return True
        if self.algorithm == "RSAPSS":
            # RSA-PSS certs can only be used by ciphers with RSA auth
            if cipher.algorithm != "RSA":
                return False
            # RSA-PSS certs do not support RSA key exchange, only RSA auth
            # "DHE" here is intended to capture both "DHE" and "ECDHE"
            if "DHE" in cipher.name:
                return True
        return False

    def compatible_with_curve(self, curve):
        if self.algorithm != "EC":
            return True
        return curve.name[-3:] == self.curve

    def compatible_with_sigalg(self, sigalg):
        if self.algorithm != sigalg.algorithm:
            return False
        sig_alg_has_curve = (
            sigalg.algorithm == "EC" and sigalg.min_protocol == Protocols.TLS13
        )
        if sig_alg_has_curve and self.curve not in sigalg.name:
            return False
        return True

    def __str__(self):
        return self.name


class Certificates(object):
    """
    When referencing certificates, use these values.
    """

    RSA_1024_SHA256 = Cert("RSA_1024_SHA256", "rsa_1024_sha256_client")
    RSA_1024_SHA384 = Cert("RSA_1024_SHA384", "rsa_1024_sha384_client")
    RSA_1024_SHA512 = Cert("RSA_1024_SHA512", "rsa_1024_sha512_client")
    RSA_2048_SHA256 = Cert("RSA_2048_SHA256", "rsa_2048_sha256_client")
    RSA_2048_SHA384 = Cert("RSA_2048_SHA384", "rsa_2048_sha384_client")
    RSA_2048_SHA512 = Cert("RSA_2048_SHA512", "rsa_2048_sha512_client")
    RSA_3072_SHA256 = Cert("RSA_3072_SHA256", "rsa_3072_sha256_client")
    RSA_3072_SHA384 = Cert("RSA_3072_SHA384", "rsa_3072_sha384_client")
    RSA_3072_SHA512 = Cert("RSA_3072_SHA512", "rsa_3072_sha512_client")
    RSA_4096_SHA256 = Cert("RSA_4096_SHA256", "rsa_4096_sha256_client")
    RSA_4096_SHA384 = Cert("RSA_4096_SHA384", "rsa_4096_sha384_client")
    RSA_4096_SHA512 = Cert("RSA_4096_SHA512", "rsa_4096_sha512_client")

    ECDSA_256 = Cert("ECDSA_256", "localhost_ecdsa_p256")
    ECDSA_384 = Cert("ECDSA_384", "ecdsa_p384_pkcs1")
    ECDSA_521 = Cert("ECDSA_521", "ecdsa_p521")

    RSA_2048_SHA256_WILDCARD = Cert(
        "RSA_2048_SHA256_WILDCARD", "rsa_2048_sha256_wildcard"
    )
    RSA_PSS_2048_SHA256 = Cert("RSA_PSS_2048_SHA256", "localhost_rsa_pss_2048_sha256")

    RSA_2048_PKCS1 = Cert("RSA_2048_PKCS1", "rsa_2048_pkcs1")

    OCSP = Cert("OCSP_RSA", "ocsp/server")
    OCSP_ECDSA = Cert("OCSP_ECDSA_256", "ocsp/server_ecdsa")


class Protocol(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __eq__(self, other):
        return other and self.value == other.value

    def __str__(self):
        return self.name


class Protocols(object):
    """
    When referencing protocols, use these protocol values.
    The first argument is the human readable name. The second
    argument is the S2N value. It is used for comparing
    protocols. Since this is hardcoded in S2N, it is not
    expected to change.
    """

    TLS13 = Protocol("TLS1.3", 34)
    TLS12 = Protocol("TLS1.2", 33)
    TLS11 = Protocol("TLS1.1", 32)
    TLS10 = Protocol("TLS1.0", 31)
    SSLv3 = Protocol("SSLv3", 30)
    SSLv2 = Protocol("SSLv2", 20)


class Cipher(object):
    def __init__(
        self,
        name,
        min_version,
        openssl1_1_1,
        fips,
        parameters=None,
        iana_standard_name=None,
        s2n=False,
        pq=False,
    ):
        self.name = name
        self.min_version = min_version
        self.openssl1_1_1 = openssl1_1_1
        self.fips = fips
        self.parameters = parameters
        self.iana_standard_name = iana_standard_name
        self.s2n = s2n
        self.pq = pq

        if self.min_version >= Protocols.TLS13:
            self.algorithm = "ANY"
        elif iana_standard_name is None:
            self.algorithm = "ANY"
        elif "ECDSA" in iana_standard_name:
            self.algorithm = "EC"
        elif "RSA" in iana_standard_name:
            self.algorithm = "RSA"
        else:
            pytest.fail("Unknown signature algorithm on cipher")

    def __eq__(self, other):
        return self.name == other

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name


class Ciphers(object):
    """
    When referencing ciphers, use these class values.
    """

    DHE_RSA_DES_CBC3_SHA = Cipher(
        "DHE-RSA-DES-CBC3-SHA",
        Protocols.SSLv3,
        False,
        False,
        iana_standard_name="SSL_DHE_RSA_WITH_3DES_EDE_CBC_SHA",
    )
    DHE_RSA_AES128_SHA = Cipher(
        "DHE-RSA-AES128-SHA",
        Protocols.SSLv3,
        True,
        False,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_128_CBC_SHA",
    )
    DHE_RSA_AES256_SHA = Cipher(
        "DHE-RSA-AES256-SHA",
        Protocols.SSLv3,
        True,
        False,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_256_CBC_SHA",
    )
    DHE_RSA_AES128_SHA256 = Cipher(
        "DHE-RSA-AES128-SHA256",
        Protocols.TLS12,
        True,
        True,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_128_CBC_SHA256",
    )
    DHE_RSA_AES256_SHA256 = Cipher(
        "DHE-RSA-AES256-SHA256",
        Protocols.TLS12,
        True,
        True,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_256_CBC_SHA256",
    )
    DHE_RSA_AES128_GCM_SHA256 = Cipher(
        "DHE-RSA-AES128-GCM-SHA256",
        Protocols.TLS12,
        True,
        True,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_128_GCM_SHA256",
    )
    DHE_RSA_AES256_GCM_SHA384 = Cipher(
        "DHE-RSA-AES256-GCM-SHA384",
        Protocols.TLS12,
        True,
        True,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_256_GCM_SHA384",
    )
    DHE_RSA_CHACHA20_POLY1305 = Cipher(
        "DHE-RSA-CHACHA20-POLY1305",
        Protocols.TLS12,
        True,
        False,
        TEST_CERT_DIRECTORY + "dhparams_2048.pem",
        iana_standard_name="TLS_DHE_RSA_WITH_AES_256_GCM_SHA384",
    )

    AES128_SHA = Cipher(
        "AES128-SHA",
        Protocols.SSLv3,
        True,
        True,
        iana_standard_name="TLS_RSA_WITH_AES_128_CBC_SHA",
    )
    AES256_SHA = Cipher(
        "AES256-SHA",
        Protocols.SSLv3,
        True,
        True,
        iana_standard_name="TLS_RSA_WITH_AES_256_CBC_SHA",
    )
    AES128_SHA256 = Cipher(
        "AES128-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_RSA_WITH_AES_128_CBC_SHA256",
    )
    AES256_SHA256 = Cipher(
        "AES256-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_RSA_WITH_AES_256_CBC_SHA256",
    )
    AES128_GCM_SHA256 = Cipher(
        "TLS_AES_128_GCM_SHA256",
        Protocols.TLS13,
        True,
        True,
        iana_standard_name="TLS_AES_128_GCM_SHA256",
    )
    AES256_GCM_SHA384 = Cipher(
        "TLS_AES_256_GCM_SHA384",
        Protocols.TLS13,
        True,
        True,
        iana_standard_name="TLS_AES_256_GCM_SHA384",
    )

    ECDHE_ECDSA_AES128_SHA = Cipher(
        "ECDHE-ECDSA-AES128-SHA",
        Protocols.SSLv3,
        True,
        False,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA",
    )
    ECDHE_ECDSA_AES256_SHA = Cipher(
        "ECDHE-ECDSA-AES256-SHA",
        Protocols.SSLv3,
        True,
        False,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA",
    )
    ECDHE_ECDSA_AES128_SHA256 = Cipher(
        "ECDHE-ECDSA-AES128-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256",
    )
    ECDHE_ECDSA_AES256_SHA384 = Cipher(
        "ECDHE-ECDSA-AES256-SHA384",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384",
    )
    ECDHE_ECDSA_AES128_GCM_SHA256 = Cipher(
        "ECDHE-ECDSA-AES128-GCM-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256",
    )
    ECDHE_ECDSA_AES256_GCM_SHA384 = Cipher(
        "ECDHE-ECDSA-AES256-GCM-SHA384",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384",
    )
    ECDHE_ECDSA_CHACHA20_POLY1305 = Cipher(
        "ECDHE-ECDSA-CHACHA20-POLY1305",
        Protocols.TLS12,
        True,
        False,
        iana_standard_name="TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256",
    )

    ECDHE_RSA_DES_CBC3_SHA = Cipher(
        "ECDHE-RSA-DES-CBC3-SHA",
        Protocols.SSLv3,
        False,
        False,
        iana_standard_name="TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA",
    )
    ECDHE_RSA_AES128_SHA = Cipher(
        "ECDHE-RSA-AES128-SHA",
        Protocols.SSLv3,
        True,
        False,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA",
    )
    ECDHE_RSA_AES256_SHA = Cipher(
        "ECDHE-RSA-AES256-SHA",
        Protocols.SSLv3,
        True,
        False,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA",
    )
    ECDHE_RSA_RC4_SHA = Cipher(
        "ECDHE-RSA-RC4-SHA",
        Protocols.SSLv3,
        False,
        False,
        iana_standard_name="TLS_ECDHE_RSA_WITH_RC4_128_SHA",
    )
    ECDHE_RSA_AES128_SHA256 = Cipher(
        "ECDHE-RSA-AES128-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256",
    )
    ECDHE_RSA_AES256_SHA384 = Cipher(
        "ECDHE-RSA-AES256-SHA384",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384",
    )
    ECDHE_RSA_AES128_GCM_SHA256 = Cipher(
        "ECDHE-RSA-AES128-GCM-SHA256",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
    )
    ECDHE_RSA_AES256_GCM_SHA384 = Cipher(
        "ECDHE-RSA-AES256-GCM-SHA384",
        Protocols.TLS12,
        True,
        True,
        iana_standard_name="TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384",
    )
    ECDHE_RSA_CHACHA20_POLY1305 = Cipher(
        "ECDHE-RSA-CHACHA20-POLY1305",
        Protocols.TLS12,
        True,
        False,
        iana_standard_name="TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256",
    )
    CHACHA20_POLY1305_SHA256 = Cipher(
        "TLS_CHACHA20_POLY1305_SHA256",
        Protocols.TLS13,
        True,
        False,
        iana_standard_name="TLS_CHACHA20_POLY1305_SHA256",
    )

    KMS_TLS_1_0_2018_10 = Cipher(
        "KMS-TLS-1-0-2018-10", Protocols.TLS10, False, False, s2n=True
    )
    PQ_TLS_1_0_2023_01 = Cipher(
        "PQ-TLS-1-0-2023-01-24", Protocols.TLS10, False, False, s2n=True, pq=True
    )
    PQ_TLS_1_3_2023_06_01 = Cipher(
        "PQ-TLS-1-3-2023-06-01", Protocols.TLS12, False, False, s2n=True, pq=True
    )

    SECURITY_POLICY_DEFAULT = Cipher(
        "default", Protocols.TLS12, False, False, s2n=True, pq=False
    )
    SECURITY_POLICY_DEFAULT_TLS13 = Cipher(
        "default_tls13", Protocols.TLS12, False, False, s2n=True, pq=False
    )
    """
    There is no "default_tls12" policy like there is a "default_tls13".
    We could use "default", but we will someday add TLS1.3 support to "default".
    Instead, pin to the latest version of "default" known to not include TLS1.3.
    See https://github.com/aws/s2n-tls/blob/main/docs/usage-guide/topics/ch06-security-policies.md#default-policy-history
    """
    SECURITY_POLICY_DEFAULT_TLS12 = Cipher(
        "20240501", Protocols.TLS12, False, False, s2n=True, pq=False
    )

    SECURITY_POLICY_20210816 = Cipher(
        "20210816", Protocols.TLS12, False, False, s2n=True, pq=False
    )

    @staticmethod
    def from_iana(iana_name):
        ciphers = [
            cipher
            for attr in vars(Ciphers)
            if not callable(cipher := getattr(Ciphers, attr))
            and not attr.startswith("_")
            and cipher.iana_standard_name
        ]
        return {cipher.iana_standard_name: cipher for cipher in ciphers}.get(iana_name)


class Curve(object):
    def __init__(self, name, min_protocol=Protocols.SSLv3):
        self.name = name
        self.min_protocol = min_protocol

    def __str__(self):
        return self.name


class Curves(object):
    """
    When referencing curves, use these class values.
    Don't hardcode curve names.
    """

    X25519 = Curve("X25519", Protocols.TLS13)
    P256 = Curve("P-256")
    # Our only SSLv3 provider doesn't support extensions
    # so there is no way to negotiate a curve other than the
    # default P-256 in SSLv3.
    P384 = Curve("P-384", Protocols.TLS10)
    P521 = Curve("P-521", Protocols.TLS10)
    SecP256r1Kyber768Draft00 = Curve("SecP256r1Kyber768Draft00")
    X25519Kyber768Draft00 = Curve("X25519Kyber768Draft00")

    @staticmethod
    def from_name(name):
        curves = [
            curve
            for attr in vars(Curves)
            if not callable(curve := getattr(Curves, attr))
            and not attr.startswith("_")
            and curve.name
        ]
        return {curve.name: curve for curve in curves}.get(name)


class KemGroup(object):
    def __init__(self, oqs_name):
        self.oqs_name = oqs_name

    def __str__(self):
        return self.oqs_name


class KemGroups(object):
    # Though s2n and oqs_openssl 3.x support KEM groups with 128-bit security
    # ECC + Kyber >512, oqs_openssl 1.1.1 does not:
    #
    # https://github.com/open-quantum-safe/openssl/blob/OQS-OpenSSL_1_1_1-stable/oqs-template/oqs-kem-info.md
    X25519_KYBER512R3 = KemGroup("X25519_kyber512")
    P256_KYBER512R3 = KemGroup("p256_kyber512")
    P384_KYBER768R3 = KemGroup("p384_kyber768")
    P521_KYBER1024R3 = KemGroup("p521_kyber1024")
    SecP256r1Kyber768Draft00 = KemGroup("SecP256r1Kyber768Draft00")
    X25519Kyber768Draft00 = KemGroup("X25519Kyber768Draft00")


class Signature(object):
    def __init__(
        self,
        name,
        min_protocol=Protocols.SSLv3,
        max_protocol=Protocols.TLS13,
        sig_type=None,
        sig_digest=None,
    ):
        self.min_protocol = min_protocol
        self.max_protocol = max_protocol

        if "RSA" in name.upper():
            self.algorithm = "RSA"
        if "PSS_PSS" in name.upper():
            self.algorithm = "RSAPSS"
        if "EC" in name.upper() or "ED" in name.upper():
            self.algorithm = "EC"

        if not (sig_type or sig_digest) and "+" in name:
            sig_type, sig_digest = name.split("+")

        self.name = name

        self.sig_type = sig_type
        self.sig_digest = sig_digest

    def __str__(self):
        return self.name


class Signatures(object):
    NONE = Signature("None+None", max_protocol=Protocols.TLS13)
    RSA_SHA1 = Signature("RSA+SHA1", max_protocol=Protocols.TLS12)
    RSA_SHA224 = Signature("RSA+SHA224", max_protocol=Protocols.TLS12)
    RSA_SHA256 = Signature("RSA+SHA256", max_protocol=Protocols.TLS12)
    RSA_SHA384 = Signature("RSA+SHA384", max_protocol=Protocols.TLS12)
    RSA_SHA512 = Signature("RSA+SHA512", max_protocol=Protocols.TLS12)
    RSA_MD5_SHA1 = Signature("RSA+MD5_SHA1", max_protocol=Protocols.TLS11)
    ECDSA_SHA224 = Signature("ECDSA+SHA224", max_protocol=Protocols.TLS12)
    ECDSA_SHA256 = Signature("ECDSA+SHA256", max_protocol=Protocols.TLS12)
    ECDSA_SHA384 = Signature("ECDSA+SHA384", max_protocol=Protocols.TLS12)
    ECDSA_SHA512 = Signature("ECDSA+SHA512", max_protocol=Protocols.TLS12)
    ECDSA_SHA1 = Signature("ECDSA+SHA1", max_protocol=Protocols.TLS12)

    RSA_PSS_RSAE_SHA256 = Signature(
        "RSA-PSS+SHA256", sig_type="RSA-PSS-RSAE", sig_digest="SHA256"
    )

    RSA_PSS_PSS_SHA256 = Signature(
        "rsa_pss_pss_sha256",
        min_protocol=Protocols.TLS12,
        sig_type="RSA-PSS-PSS",
        sig_digest="SHA256",
    )

    ECDSA_SECP256r1_SHA256 = Signature(
        "ecdsa_secp256r1_sha256",
        min_protocol=Protocols.TLS13,
        sig_type="ECDSA",
        sig_digest="SHA256",
    )
    ECDSA_SECP384r1_SHA384 = Signature(
        "ecdsa_secp384r1_sha384",
        min_protocol=Protocols.TLS13,
        sig_type="ECDSA",
        sig_digest="SHA384",
    )
    ECDSA_SECP521r1_SHA512 = Signature(
        "ecdsa_secp521r1_sha512",
        min_protocol=Protocols.TLS13,
        sig_type="ECDSA",
        sig_digest="SHA512",
    )


class Results(object):
    """
    An instance of this object will be returned to the test by a managed_process'
    get_results() method.
    """

    # Byte array containing the standard output of the process
    stdout = None

    # Byte array containing the standard error of the process
    stderr = None

    # Exit code of the process
    exit_code = None

    # Any exception thrown while running the process
    exception = None

    def __init__(
        self,
        stdout,
        stderr,
        exit_code,
        exception,
        expect_stderr=False,
        expect_nonzero_exit=False,
    ):
        self.stdout = stdout
        self.stderr = stderr
        self.exit_code = exit_code
        self.exception = exception
        self.expect_stderr = expect_stderr
        self.expect_nonzero_exit = expect_nonzero_exit

    def __str__(self):
        return "Stdout: {}\nStderr: {}\nExit code: {}\nException: {}".format(
            self.stdout, self.stderr, self.exit_code, self.exception
        )

    def assert_success(self):
        assert self.exception is None, self.exception
        if not self.expect_nonzero_exit:
            assert self.exit_code == 0, f"exit code: {self.exit_code}"
        if not self.expect_stderr:
            assert not self.stderr, self.stderr

    def output_streams(self):
        return {self.stdout, self.stderr}


class ProviderOptions(object):
    def __init__(
        self,
        mode=None,
        host=None,
        port=None,
        cipher=None,
        curve=None,
        key=None,
        cert=None,
        use_session_ticket=False,
        insecure=False,
        data_to_send=None,
        use_client_auth=False,
        extra_flags=None,
        trust_store=None,
        reconnects_before_exit=None,
        reconnect=None,
        verify_hostname=None,
        server_name=None,
        protocol=None,
        use_mainline_version=None,
        env_overrides=dict(),
        enable_client_ocsp=False,
        ocsp_response=None,
        signature_algorithm=None,
        record_size=None,
        verbose=None,
        echo=True,
    ):
        # Client or server
        self.mode = mode

        # Hostname
        self.host = host
        if not self.host:
            self.host = "localhost"

        # Port (string because this will be converted to a command line
        self.port = str(port)

        # Cipher suite
        self.cipher = cipher

        # Named curve
        self.curve = curve

        # Path to a key PEM
        self.key = key

        # Path to a certificate PEM
        self.cert = cert

        self.trust_store = trust_store

        # Boolean whether to use a resumption ticket
        self.use_session_ticket = use_session_ticket

        # Boolean whether to disable x509 verification
        self.insecure = insecure

        # Which protocol to use with this provider
        self.protocol = protocol

        # This data will be sent to the peer
        self.data_to_send = data_to_send

        # Parameters to configure client authentication
        self.use_client_auth = use_client_auth

        # Reconnects on the server side (includes first connection)
        self.reconnects_before_exit = reconnects_before_exit

        # Tell the client to reconnect
        self.reconnect = reconnect

        # Tell the client to verify that the hostname returned by the server
        # matches this argument
        self.verify_hostname = verify_hostname

        # Tell the client to send this server name to the server
        self.server_name = server_name

        # Extra flags to pass to the provider
        self.extra_flags = extra_flags

        # Boolean whether the provider is an older version of s2n
        self.use_mainline_version = use_mainline_version

        # Extra environment parameters
        self.env_overrides = env_overrides

        # Enable OCSP on the client
        self.enable_client_ocsp = enable_client_ocsp

        # Path to OCSP response on the server
        self.ocsp_response = ocsp_response

        self.signature_algorithm = signature_algorithm

        self.record_size = record_size

        # How verbose should the provider be when printing to stdout?
        # Default to more information, leave the option for less.
        # Useful if you find that debugging information is printed between
        # application data you expect the provider to print on stdout.
        self.verbose = verbose

        # Should the provider echo any data sent / received?
        # If not, the connection will be closed after one handshake.
        self.echo = echo



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from io import BytesIO
import unittest
import os
import tempfile
import math
import shutil
import time
from test import NativeResourceTest
from concurrent.futures import Future
from multiprocessing import Process
import multiprocessing as mp
import sys
import gc

from awscrt.http import HttpHeaders, HttpRequest
from awscrt.auth import AwsCredentials
from awscrt.s3 import (
    S3ChecksumAlgorithm,
    S3ChecksumConfig,
    S3ChecksumLocation,
    S3Client,
    S3RequestType,
    S3ResponseError,
    CrossProcessLock,
    S3FileIoOptions,
    create_default_s3_signing_config,
    get_optimized_platforms,
)
from awscrt.io import (
    ClientBootstrap,
    ClientTlsContext,
    DefaultHostResolver,
    EventLoopGroup,
    TlsConnectionOptions,
    TlsContextOptions,
)
from awscrt.auth import (
    AwsCredentialsProvider,
    AwsSignatureType,
    AwsSignedBodyHeaderType,
    AwsSignedBodyValue,
    AwsSigningAlgorithm,
    AwsSigningConfig,
)
from awscrt.common import join_all_native_threads

MB = 1024 ** 2
GB = 1024 ** 3
S3EXPRESS_ENDPOINT = "crts-east1--use1-az4--x-s3.s3express-use1-az4.us-east-1.amazonaws.com"

cross_process_lock_name = "instance_lock_test"


def cross_proc_task():
    try:
        lock = CrossProcessLock(cross_process_lock_name)
        lock.acquire()
        lock.release()
        exit(0)
    except RuntimeError as e:
        exit(-1)


def release_lock_task():
    # remove the global lock
    global CRT_S3_PROCESS_LOCK
    CRT_S3_PROCESS_LOCK = None
    exit(0)


class CrossProcessLockTest(NativeResourceTest):

    def setUp(self):
        self.nonce = time.time()
        super().setUp()

    def test_with_statement(self):
        nonce_str = f'lock_a_{self.nonce}'
        with CrossProcessLock(nonce_str) as lock:
            try:
                new_lock = CrossProcessLock(nonce_str)
                new_lock.acquire()
                self.fail("Acquiring a lock by the same nonce should fail when it's already held")
            except RuntimeError as e:
                unique_nonce_str = f'lock_b{self.nonce}'
                new_lock = CrossProcessLock(unique_nonce_str)
                new_lock.acquire()
                new_lock.release()

        lock_after_with_same_nonce = CrossProcessLock(nonce_str)
        lock_after_with_same_nonce.acquire()
        lock_after_with_same_nonce.release()

    def test_cross_proc(self):
        with CrossProcessLock(cross_process_lock_name) as lock:
            process = Process(target=cross_proc_task)
            process.start()
            process.join()
            # acquiring this lock in a sub-process should fail since we
            # already hold the lock in this process.
            self.assertNotEqual(0, process.exitcode)
        # now that we've released the lock above, the same sub-process path
        # should now succeed.
        unlocked_process = Process(target=cross_proc_task)
        unlocked_process.start()
        unlocked_process.join()
        self.assertEqual(0, unlocked_process.exitcode)

    @unittest.skipIf(sys.platform.startswith('win'), "Windows doesn't support fork")
    def test_fork_shares_lock(self):
        # mock the use case from boto3 where a global lock used and the workaround with fork.
        global CRT_S3_PROCESS_LOCK
        CRT_S3_PROCESS_LOCK = CrossProcessLock(cross_process_lock_name)
        CRT_S3_PROCESS_LOCK.acquire()
        mp.set_start_method('fork', force=True)
        # the first forked process release the forked lock.
        # when the process forked, the child process also has the lock and it could release the lock before
        # the parent process. Make sure when this happens, the lock is still held by the parent process.
        release_process = Process(target=release_lock_task)
        release_process.start()
        release_process.join()

        # create another process try to acquire the lock, it should fail as the
        # lock should still be held with the main process.
        process = Process(target=cross_proc_task)
        process.start()
        process.join()
        # acquiring this lock in a sub-process should fail since we
        # already hold the lock in this process.
        self.assertNotEqual(0, process.exitcode)
        del CRT_S3_PROCESS_LOCK


class FileCreator(object):
    def __init__(self):
        self.rootdir = tempfile.mkdtemp()

    def remove_all(self):
        shutil.rmtree(self.rootdir)

    def create_file(self, filename, contents, mode='w'):
        """Creates a file in a tmpdir
        ``filename`` should be a relative path, e.g. "foo/bar/baz.txt"
        It will be translated into a full path in a tmp dir.
        ``mode`` is the mode the file should be opened either as ``w`` or
        `wb``.
        Returns the full path to the file.
        """
        full_path = os.path.join(self.rootdir, filename)
        if not os.path.isdir(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))
        with open(full_path, mode) as f:
            f.write(contents)
        return full_path

    def create_file_with_size(self, filename, filesize):
        filename = self.create_file(filename, contents='')
        chunksize = 8192
        with open(filename, 'wb') as f:
            for i in range(int(math.ceil(filesize / float(chunksize)))):
                f.write(b'a' * chunksize)
        return filename

    def append_file(self, filename, contents):
        """Append contents to a file
        ``filename`` should be a relative path, e.g. "foo/bar/baz.txt"
        It will be translated into a full path in a tmp dir.
        Returns the full path to the file.
        """
        full_path = os.path.join(self.rootdir, filename)
        if not os.path.isdir(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))
        with open(full_path, 'a') as f:
            f.write(contents)
        return full_path

    def full_path(self, filename):
        """Translate relative path to full path in temp dir.
        f.full_path('foo/bar.txt') -> /tmp/asdfasd/foo/bar.txt
        """
        return os.path.join(self.rootdir, filename)


def s3_client_new(
        secure,
        region,
        part_size=0,
        is_cancel_test=False,
        enable_s3express=False,
        mem_limit=None,
        network_interface_names=None):

    if is_cancel_test:
        # for cancellation tests, make things slow, so it's less likely that
        # stuff succeeds on other threads before the cancellation is processed.
        num_threads = 1
        throughput_target_gbps = 0.000028  # 28 Kbps beeepdiiingeep beeeeeekskhskshhKKKKchCH
    else:
        # else use defaults
        num_threads = None
        throughput_target_gbps = None

    event_loop_group = EventLoopGroup(num_threads)
    host_resolver = DefaultHostResolver(event_loop_group)
    bootstrap = ClientBootstrap(event_loop_group, host_resolver)
    credential_provider = AwsCredentialsProvider.new_default_chain(bootstrap)
    signing_config = create_default_s3_signing_config(region=region, credential_provider=credential_provider)
    tls_option = None
    if secure:
        opt = TlsContextOptions()
        ctx = ClientTlsContext(opt)
        tls_option = TlsConnectionOptions(ctx)
    s3_client = S3Client(
        bootstrap=bootstrap,
        region=region,
        signing_config=signing_config,
        tls_connection_options=tls_option,
        part_size=part_size,
        throughput_target_gbps=throughput_target_gbps,
        enable_s3express=enable_s3express,
        memory_limit=mem_limit,
        network_interface_names=network_interface_names)
    return s3_client


class FakeReadStream(object):
    def __init__(self, read_future):
        self._future = read_future

    def read(self, length):
        fake_string = "x" * length
        fake_data = bytes(fake_string, 'utf-8')
        if not self._future.done():
            self._future.set_result(None)
        return fake_data


class S3ClientTest(NativeResourceTest):

    def setUp(self):
        self.region = "us-west-2"
        self.timeout = 10  # seconds
        super().setUp()

    def test_sanity(self):
        s3_client = s3_client_new(False, self.region)
        self.assertIsNotNone(s3_client)

    def test_sanity_secure(self):
        s3_client = s3_client_new(True, self.region)
        self.assertIsNotNone(s3_client)

    def test_sanity_network_interface_names(self):
        # This is just a sanity test to ensure that we are passing the parameter correctly.
        with self.assertRaises(Exception):
            s3_client_new(True, self.region, network_interface_names=("eth0", "invalid-network-interface"))

    def test_wait_shutdown(self):
        s3_client = s3_client_new(False, self.region)
        self.assertIsNotNone(s3_client)

        shutdown_event = s3_client.shutdown_event
        del s3_client
        self.assertTrue(shutdown_event.wait(self.timeout))

    def test_get_optimized_platforms(self):
        platform_list = get_optimized_platforms()
        self.assertTrue(len(platform_list) > 0)
        self.assertTrue("p4d.24xlarge" in platform_list)


@unittest.skipUnless(os.environ.get('AWS_TEST_S3'), 'set env var to run test: AWS_TEST_S3')
class S3RequestTest(NativeResourceTest):
    def setUp(self):
        super().setUp()
        # TODO: use env-vars to customize how these tests are run, instead of relying on hard-coded values
        self.get_test_object_path = "/get_object_test_10MB.txt"
        self.put_test_object_path = "/put_object_test_py_10MB.txt"
        self.region = "us-west-2"
        self.bucket_name = "aws-crt-canary-bucket"
        self.timeout = 100  # seconds
        self.num_threads = 0
        self.special_path = "put_object_test_10MB@$%.txt"
        self.non_ascii_file_name = "ÉxÅmple.txt"
        self.part_size = 5 * MB

        self.response_headers = None
        self.response_status_code = None
        self.received_body_len = 0
        self.transferred_len = 0
        self.data_len = 0
        self.progress_invoked = 0
        self.done_error = None
        self.done_status_code = None
        self.done_error_headers = None
        self.done_error_body = None
        self.done_error_operation_name = None
        self.done_did_validate_checksum = None
        self.done_checksum_validation_algorithm = None

        self.files = FileCreator()
        self.temp_put_obj_file_path = self.files.create_file_with_size("temp_put_obj_10mb", 10 * MB)
        self.s3express_preload_cache = [('key_1', AwsCredentials("accesskey_1", "secretAccessKey", "sessionToken")),
                                        ('key_2', AwsCredentials("accesskey_2", "secretAccessKey", "sessionToken"))]

    def tearDown(self):
        self.files.remove_all()
        self.s3express_preload_cache = None
        super().tearDown()

    def _build_endpoint_string(self, region, bucket_name, enable_s3express=False):
        if enable_s3express:
            return S3EXPRESS_ENDPOINT
        return bucket_name + ".s3." + region + ".amazonaws.com"

    def _get_object_request(self, object_path, enable_s3express=False):
        headers = HttpHeaders([("host", self._build_endpoint_string(self.region, self.bucket_name, enable_s3express))])
        request = HttpRequest("GET", object_path, headers)
        return request

    def _put_object_request(
            self,
            input_stream,
            content_len,
            path=None,
            unknown_content_length=False,
            enable_s3express=False):
        # if send file path is set, the body_stream of http request will be ignored (using file handler from C instead)
        headers = HttpHeaders([("host", self._build_endpoint_string(self.region, self.bucket_name, enable_s3express)),
                               ("Content-Type", "text/plain")])
        if unknown_content_length is False:
            headers.add("Content-Length", str(content_len))
        if path is None:
            path = self.put_test_object_path
        request = HttpRequest("PUT", path, headers, input_stream)
        return request

    def _on_request_headers(self, status_code, headers, **kargs):
        self.response_status_code = status_code
        self.response_headers = headers

    def _on_request_body(self, chunk, offset, **kargs):
        self.received_body_len = self.received_body_len + len(chunk)

    def _on_request_done(
            self,
            error,
            error_headers,
            error_body,
            error_operation_name,
            status_code,
            did_validate_checksum,
            checksum_validation_algorithm,
            **kwargs):
        self.done_error = error
        self.done_error_headers = error_headers
        self.done_error_body = error_body
        self.done_error_operation_name = error_operation_name
        self.done_status_code = status_code
        self.done_did_validate_checksum = did_validate_checksum
        self.done_checksum_validation_algorithm = checksum_validation_algorithm

    def _on_progress(self, progress):
        self.transferred_len += progress

    def _validate_successful_response(self, is_put_object):
        self.assertEqual(self.response_status_code, 200, "status code is not 200")
        self.assertEqual(self.done_status_code, self.response_status_code,
                         "status-code from on_done doesn't match code from on_headers")
        self.assertIsNone(self.done_error)
        self.assertIsNone(self.done_error_headers)
        self.assertIsNone(self.done_error_body)
        self.assertIsNone(self.done_error_operation_name)
        self.assertIsInstance(self.done_did_validate_checksum, bool)
        if self.done_did_validate_checksum:
            self.assertIsInstance(self.done_checksum_validation_algorithm, S3ChecksumAlgorithm)
        else:
            self.assertIsNone(self.done_checksum_validation_algorithm)
        headers = HttpHeaders(self.response_headers)
        self.assertIsNone(headers.get("Content-Range"))
        body_length = headers.get("Content-Length")
        if not is_put_object:
            self.assertIsNotNone(body_length, "Content-Length is missing from headers")
        if body_length:
            self.assertEqual(
                int(body_length),
                self.received_body_len,
                "Received body length does not match the Content-Length header")

    def _test_s3_put_get_object(
            self,
            request,
            request_type,
            exception_name=None,
            enable_s3express=False,
            mem_limit=None,
            **kwargs):
        s3_client = s3_client_new(
            False,
            self.region,
            self.part_size,
            enable_s3express=enable_s3express,
            mem_limit=mem_limit)
        signing_config = None
        if enable_s3express:
            signing_config = AwsSigningConfig(
                algorithm=AwsSigningAlgorithm.V4_S3EXPRESS)

        s3_request = s3_client.make_request(
            request=request,
            type=request_type,
            signing_config=signing_config,
            on_headers=self._on_request_headers,
            on_body=self._on_request_body,
            on_done=self._on_request_done,
            **kwargs)

        finished_future = s3_request.finished_future
        shutdown_event = s3_request.shutdown_event
        s3_request = None
        self.assertTrue(shutdown_event.wait(self.timeout))

        if exception_name is None:
            try:
                finished_future.result()
                self._validate_successful_response(request_type is S3RequestType.PUT_OBJECT)
            except S3ResponseError as e:
                print(e.status_code)
                print(e.headers)
                print(e.body)
                raise e
        else:
            e = finished_future.exception()
            self.assertEqual(e.name, exception_name)
            self.assertEqual(e, self.done_error)

    def test_get_object(self):
        request = self._get_object_request(self.get_test_object_path)
        self._test_s3_put_get_object(request, S3RequestType.GET_OBJECT)

    def test_get_object_mem_limit(self):
        request = self._get_object_request(self.get_test_object_path)
        self._test_s3_put_get_object(request, S3RequestType.GET_OBJECT, mem_limit=2 * GB)

    def test_put_object(self):
        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(put_body_stream, content_length)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT)
        put_body_stream.close()

    def test_put_object_mem_limit(self):
        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(put_body_stream, content_length)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT, mem_limit=2 * GB)
        put_body_stream.close()

    def test_put_object_unknown_content_length(self):
        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(put_body_stream, content_length, unknown_content_length=True)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT)
        put_body_stream.close()

    def test_put_object_unknown_content_length_single_part(self):
        data_bytes = "test crt python single part upload".encode(encoding='utf-8')
        put_body_stream = BytesIO(data_bytes)
        request = self._put_object_request(put_body_stream, len(data_bytes), unknown_content_length=True)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT)
        put_body_stream.close()

    def test_get_object_s3express(self):
        self.region = "us-east-1"
        request = self._get_object_request("/crt-download-10MB", enable_s3express=True)
        self._test_s3_put_get_object(request, S3RequestType.GET_OBJECT, enable_s3express=True)

    def test_put_object_s3express(self):
        self.region = "us-east-1"
        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(put_body_stream, content_length, enable_s3express=True)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT, enable_s3express=True)
        put_body_stream.close()

    def test_put_object_multiple_times(self):
        s3_client = s3_client_new(False, self.region, 5 * MB)
        finished_futures = []
        for i in range(3):
            tempfile = self.files.create_file_with_size("temp_file_{}".format(str(i)), 10 * MB)
            path = "/put_object_test_py_10MB_{}.txt".format(str(i))
            content_length = os.stat(tempfile).st_size
            request = self._put_object_request(None, content_length, path=path)
            s3_request = s3_client.make_request(
                request=request,
                type=S3RequestType.PUT_OBJECT,
                send_filepath=tempfile,
                on_headers=self._on_request_headers,
                on_body=self._on_request_body,
                on_done=self._on_request_done)
            finished_futures.append(s3_request.finished_future)
            # request keeps connection alive. delete pointer so connection can shut down
            del s3_request
        try:
            for future in finished_futures:
                future.result(self.timeout)
        except Exception as e:
            # failed
            self.assertTrue(False)

        client_shutdown_event = s3_client.shutdown_event
        del s3_client
        self.assertTrue(client_shutdown_event.wait(self.timeout))

    def test_put_object_request_override_part_size(self):
        s3_client = s3_client_new(False, self.region, 5 * MB)

        tempfile = self.files.create_file_with_size("temp_file_override", 10 * MB)
        path = "/put_object_test_py_10MB_override.txt"
        content_length = os.stat(tempfile).st_size
        request = self._put_object_request(None, content_length, path=path)
        # Override the threshold to 10 MB, which will result in a single part upload
        s3_request = s3_client.make_request(
            request=request,
            type=S3RequestType.PUT_OBJECT,
            send_filepath=tempfile,
            on_headers=self._on_request_headers,
            on_body=self._on_request_body,
            on_done=self._on_request_done,
            multipart_upload_threshold=10 * MB)
        try:
            s3_request.finished_future.result(self.timeout)
        except Exception as e:
            # failed
            self.assertTrue(False)

        # Etag headers for a MPU will be formatted with `-[part number]`
        etag = HttpHeaders(self.response_headers).get("Etag")
        # make sure we uploaded as single part as we override the threshold
        self.assertFalse("-" in etag)

        del s3_request
        client_shutdown_event = s3_client.shutdown_event
        del s3_client
        self.assertTrue(client_shutdown_event.wait(self.timeout))

    def test_get_object_filepath(self):
        request = self._get_object_request(self.get_test_object_path)
        request_type = S3RequestType.GET_OBJECT
        s3_client = s3_client_new(False, self.region, 5 * MB)
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as file:
            file.close()
            s3_request = s3_client.make_request(
                request=request,
                type=request_type,
                recv_filepath=file.name,
                on_headers=self._on_request_headers,
                on_progress=self._on_progress,
                on_done=self._on_request_done)
            finished_future = s3_request.finished_future

            # Regression test: Let S3Request get GC'd early.
            # The download should continue without problems.
            # We once had a bug where the file would get closed too early:
            # https://github.com/awslabs/aws-crt-python/pull/506
            del s3_request

            finished_future.result(self.timeout)

            # Result check
            self.data_len = int(HttpHeaders(self.response_headers).get("Content-Length"))
            file_stats = os.stat(file.name)
            file_len = file_stats.st_size
            self.assertEqual(
                file_len,
                self.transferred_len,
                "the length of written file does not match the transferred length reported")
            self.assertEqual(
                self.data_len,
                self.transferred_len,
                "the transferred length reported does not match the content-length header")
            self.assertEqual(self.response_status_code, 200, "status code is not 200")
            # TODO verify the content of written file
            os.remove(file.name)

    def test_put_object_filepath(self):
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(None, content_length)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT, send_filepath=self.temp_put_obj_file_path)

    def test_put_object_filepath_with_fio_options(self):
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(None, content_length)
        fio_options = S3FileIoOptions(should_stream=True, disk_throughput_gbps=10.0, direct_io=True)
        self._test_s3_put_get_object(
            request,
            S3RequestType.PUT_OBJECT,
            send_filepath=self.temp_put_obj_file_path,
            fio_options=fio_options)

    def test_put_object_filepath_unknown_content_length(self):
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(None, content_length, unknown_content_length=True)
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT, send_filepath=self.temp_put_obj_file_path)

    def test_put_object_filepath_move(self):
        # remove the input file when request done
        tempfile = self.files.create_file_with_size("temp_file", 10 * MB)
        content_length = os.stat(tempfile).st_size
        request = self._put_object_request(None, content_length)
        s3_client = s3_client_new(False, self.region, 5 * MB)
        request_type = S3RequestType.PUT_OBJECT
        done_future = Future()

        def on_done_remove_file(**kwargs):
            self._on_request_done(**kwargs)
            os.remove(tempfile)
            done_future.set_result(None)

        s3_client.make_request(
            request=request,
            type=request_type,
            send_filepath=tempfile,
            on_headers=self._on_request_headers,
            on_progress=self._on_progress,
            on_done=on_done_remove_file)
        done_future.result(self.timeout)

        # check result
        self.assertEqual(
            content_length,
            self.transferred_len,
            "the transferred length reported does not match body we sent")
        self._validate_successful_response(request_type is S3RequestType.PUT_OBJECT)

    def upload_with_global_client(self):
        global CRT_S3_CLIENT
        if CRT_S3_CLIENT is None:
            CRT_S3_CLIENT = s3_client_new(False, self.region, 5 * MB)
        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        request = self._put_object_request(put_body_stream, content_length)
        s3_request = CRT_S3_CLIENT.make_request(
            request=request,
            type=S3RequestType.PUT_OBJECT,
            on_headers=self._on_request_headers,
            on_body=self._on_request_body,
            on_done=self._on_request_done)

        shutdown_event = s3_request.shutdown_event
        s3_request = None
        self.assertTrue(shutdown_event.wait(self.timeout))
        put_body_stream.close()

    def fork_s3_client(self):
        try:
            # init_logging(LogLevel.Trace, f"aws-crt-python{os.getpid()}.log")
            self.upload_with_global_client()
            global CRT_S3_CLIENT
            del CRT_S3_CLIENT
            exit(0)
        except Exception as e:
            print(f"fork_s3_client error: {e}")
            exit(-1)

    def before_fork(self):
        global CRT_S3_CLIENT
        try:
            if CRT_S3_CLIENT is not None:
                # The client is not safe to use after fork, so we need to release it.
                # make sure the client is shutdown properly before fork
                # also wait for every thread to be joined, incase of some thread is in the middle of cleanup.
                shutdown_event = CRT_S3_CLIENT.shutdown_event
                CRT_S3_CLIENT = None
                gc.collect()
                self.assertTrue(shutdown_event.wait(self.timeout))
                self.assertTrue(join_all_native_threads(timeout_sec=10))
        except Exception as e:
            print(f"before_fork error: {e}")
            # fail hard as exceptions raised from the fork handler will be ignored.
            sys.stdout.flush()
            sys.stderr.flush()
            os._exit(-1)

    @unittest.skipIf(sys.platform.startswith('win') or sys.platform == 'darwin',
                     "Test skipped on Windows and macOS. Windows doesn't support fork. macOS has background threads crashes the forked process.")
    def test_fork_workaround(self):
        # mock the boto3 use case where a global client is used and the
        # workaround for fork is to release the client from the fork handler.
        global CRT_S3_CLIENT
        CRT_S3_CLIENT = s3_client_new(False, self.region, 5 * MB)
        self.upload_with_global_client()
        # fork the process and use the global S3 client for the transfer.
        # to workaround the background thread issue that cleaned after fork,
        # we need to make sure the client is shutdown properly
        # and all background threads has joined before fork, see `self.before_fork`
        # And in the sub-process, we can recreate the global S3 client and use it.
        os.register_at_fork(before=self.before_fork)
        mp.set_start_method('fork', force=True)
        process = Process(target=self.fork_s3_client)
        process.start()
        process.join()
        self.assertEqual(0, process.exitcode)
        self.upload_with_global_client()
        del CRT_S3_CLIENT

    def _round_trip_with_checksums_helper(
            self,
            algo=S3ChecksumAlgorithm.CRC32,
            mpu=True,
            provide_full_object_checksum=False):
        if not mpu:
            # increase the part size for the client to use single part upload
            self.part_size = 20 * MB

        put_body_stream = open(self.temp_put_obj_file_path, "rb")
        content_length = os.stat(self.temp_put_obj_file_path).st_size
        # construct different path to prevent race condition between tests
        path = '/hello-world-' + algo.name
        if mpu:
            path += "-mpu"
        if provide_full_object_checksum:
            path += "-full-object"

        if algo == S3ChecksumAlgorithm.CRC32:
            checksum_header_name = 'x-amz-checksum-crc32'
            checksum_str = 'a9ccsg=='
        elif algo == S3ChecksumAlgorithm.CRC64NVME:
            checksum_header_name = 'x-amz-checksum-crc64nvme'
            checksum_str = 'tPMvgM0jSDQ='
        else:
            raise Exception("Checksum algo not supported by test helper")

        # upload, with client adding checksum
        upload_request = self._put_object_request(put_body_stream, content_length, path=path)
        upload_checksum_config = S3ChecksumConfig(
            algorithm=algo,
            location=S3ChecksumLocation.TRAILER)
        if provide_full_object_checksum:
            upload_request.headers.add(checksum_header_name, checksum_str)
            # checksum will be provided from the header, don't set the checksum configs
            upload_checksum_config = None

        self._test_s3_put_get_object(upload_request, S3RequestType.PUT_OBJECT,
                                     checksum_config=upload_checksum_config)

        # download, with client validating checksum
        download_request = self._get_object_request(path)
        download_checksum_config = S3ChecksumConfig(validate_response=True)
        self._test_s3_put_get_object(download_request, S3RequestType.GET_OBJECT,
                                     checksum_config=download_checksum_config)
        self.assertTrue(self.done_did_validate_checksum)
        self.assertEqual(self.done_checksum_validation_algorithm, algo)
        self.assertEqual(HttpHeaders(self.response_headers).get(checksum_header_name),
                         checksum_str)
        put_body_stream.close()

    def test_round_trip_with_trailing_checksum(self):
        self._round_trip_with_checksums_helper(S3ChecksumAlgorithm.CRC32, mpu=False)

    def test_round_trip_with_full_object_checksum_mpu(self):
        self._round_trip_with_checksums_helper(
            S3ChecksumAlgorithm.CRC64NVME,
            mpu=True,
            provide_full_object_checksum=True)

    def test_round_trip_with_full_object_checksum_single_part(self):
        self._round_trip_with_checksums_helper(
            S3ChecksumAlgorithm.CRC64NVME,
            mpu=False,
            provide_full_object_checksum=True)

    def test_round_trip_with_full_object_checksum_mpu_crc32(self):
        self._round_trip_with_checksums_helper(S3ChecksumAlgorithm.CRC32, mpu=True, provide_full_object_checksum=True)

    def test_round_trip_with_full_object_checksum_single_part_crc32(self):
        self._round_trip_with_checksums_helper(S3ChecksumAlgorithm.CRC32, mpu=False, provide_full_object_checksum=True)

    def _on_progress_cancel_after_first_chunk(self, progress):
        self.transferred_len += progress
        self.progress_invoked += 1
        self.s3_request.cancel()

    def test_multipart_get_object_cancel(self):
        # a 5 GB file
        request = self._get_object_request("/get_object_test_5120MB.txt")
        s3_client = s3_client_new(False, self.region, 5 * MB, is_cancel_test=True)
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as file:
            file.close()
            self.s3_request = s3_client.make_request(
                request=request,
                recv_filepath=file.name,
                type=S3RequestType.GET_OBJECT,
                on_headers=self._on_request_headers,
                on_progress=self._on_progress_cancel_after_first_chunk,
                on_done=self._on_request_done)
            finished_future = self.s3_request.finished_future
            e = finished_future.exception(self.timeout)
            self.assertEqual(e.name, "AWS_ERROR_S3_CANCELED")

            # Result check
            self.data_len = int(HttpHeaders(self.response_headers).get("Content-Length"))
            self.assertLess(
                self.transferred_len,
                self.data_len,
                "the cancel failed to block all the following body")

            # The on_finish callback may invoke the progress
            self.assertLessEqual(self.progress_invoked, 2)
            shutdown_event = self.s3_request.shutdown_event
            self.s3_request = None
            self.assertTrue(shutdown_event.wait(self.timeout))
            os.remove(file.name)

    def test_get_object_quick_cancel(self):
        # a 5 GB file
        request = self._get_object_request("/get_object_test_5120MB.txt")
        s3_client = s3_client_new(False, self.region, 5 * MB, is_cancel_test=True)
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as file:
            file.close()
            s3_request = s3_client.make_request(
                request=request,
                recv_filepath=file.name,
                type=S3RequestType.GET_OBJECT,
                on_headers=self._on_request_headers,
                on_progress=self._on_progress,
                on_done=self._on_request_done)
            s3_request.cancel()
            finished_future = s3_request.finished_future
            e = finished_future.exception(self.timeout)
            self.assertEqual(e.name, "AWS_ERROR_S3_CANCELED")
            shutdown_event = s3_request.shutdown_event
            s3_request = None
            self.assertTrue(shutdown_event.wait(self.timeout))
            os.remove(file.name)

    def _put_object_cancel_helper(self, cancel_after_read):
        read_future = Future()
        put_body_stream = FakeReadStream(read_future)
        data_len = 10 * GB  # some fake length
        headers = HttpHeaders([("host", self._build_endpoint_string(self.region, self.bucket_name)),
                               ("Content-Type", "text/plain"), ("Content-Length", str(data_len))])
        http_request = HttpRequest("PUT", "/cancelled_request", headers, put_body_stream)
        s3_client = s3_client_new(False, self.region, 8 * MB, is_cancel_test=True)
        s3_request = s3_client.make_request(
            request=http_request,
            type=S3RequestType.PUT_OBJECT,
            on_headers=self._on_request_headers,
            on_done=self._on_request_done)

        if cancel_after_read:
            read_future.result(self.timeout)
        s3_request.cancel()
        finished_future = s3_request.finished_future
        e = finished_future.exception(self.timeout)
        self.assertEqual(e.name, "AWS_ERROR_S3_CANCELED")

        shutdown_event = s3_request.shutdown_event
        s3_request = None
        self.assertTrue(shutdown_event.wait(self.timeout))
        # TODO If CLI installed, run the following command to ensure the cancel succeed.
        # aws s3api list-multipart-uploads --bucket aws-crt-canary-bucket --prefix 'cancelled_request'
        # Nothing should printout

    def test_multipart_put_object_cancel(self):
        return self._put_object_cancel_helper(True)

    def test_put_object_quick_cancel(self):
        return self._put_object_cancel_helper(False)

    def test_singlepart_upload_with_invalid_request(self):
        # send upload with incorrect Content-MD5
        # need to do single-part upload so the Content-MD5 header is sent along as-is.
        content_length = 100
        file_path = self.files.create_file_with_size("temp_file", content_length)
        put_body_stream = open(file_path, "r+b")
        request = self._put_object_request(put_body_stream, content_length)
        request.headers.set("Content-MD5", "something")
        self._test_s3_put_get_object(request, S3RequestType.PUT_OBJECT, "AWS_ERROR_S3_INVALID_RESPONSE_STATUS")

        # check that data from on_done callback came through correctly
        self.assertIsInstance(self.done_error, S3ResponseError)
        self.assertEqual(self.done_status_code, 400)
        self.assertEqual(self.done_error.status_code, 400)
        self.assertIsNotNone(self.done_error_headers)
        self.assertTrue(any(h[0].lower() == 'x-amz-request-id' for h in self.done_error_headers))
        self.assertListEqual(self.done_error_headers, self.done_error.headers)
        self.assertIsNotNone(self.done_error_body)
        self.assertEqual(self.done_error_operation_name, "PutObject")
        self.assertEqual(self.done_error_operation_name, self.done_error.operation_name)
        self.assertTrue(b"InvalidDigest" in self.done_error_body)
        self.assertEqual(self.done_error_body, self.done_error.body)

        put_body_stream.close()

    def test_default_request_failure(self):
        # send invalid DEFAULT S3Request
        # ensure error info (including custom operation_name) comes through correctly
        s3_client = S3Client(region=self.region)

        # send invalid request to S3.
        http_request = HttpRequest(method="GET", path="/obviously-invalid-path-object-does-not-exist")
        http_request.headers.add("host", self._build_endpoint_string(self.region, self.bucket_name))
        http_request.headers.add("content-length", "0")
        s3_request = s3_client.make_request(
            type=S3RequestType.DEFAULT,
            request=http_request,
            operation_name="MyNewOperationName")

        exception = s3_request.finished_future.exception(self.timeout)
        self.assertIsInstance(exception, S3ResponseError)
        self.assertEqual(exception.operation_name, "MyNewOperationName")

    def test_on_headers_callback_failure(self):
        def _explode(**kwargs):
            raise RuntimeError("Error in on_headers callback")

        request = self._get_object_request(self.get_test_object_path)
        s3_client = s3_client_new(False, self.region, 5 * MB)
        s3_request = s3_client.make_request(
            request=request,
            type=S3RequestType.GET_OBJECT,
            on_headers=_explode,
            on_body=self._on_request_body,
        )

        finished_future = s3_request.finished_future
        shutdown_event = s3_request.shutdown_event
        s3_request = None
        self.assertTrue(shutdown_event.wait(self.timeout))

        e = finished_future.exception()
        # check that data from on_done callback came through correctly
        self.assertIsInstance(e, RuntimeError)
        self.assertEqual(str(e), "Error in on_headers callback")

    def test_on_body_callback_failure(self):
        def _explode(**kwargs):
            raise RuntimeError("Error in on_body callback")

        request = self._get_object_request(self.get_test_object_path)
        s3_client = s3_client_new(False, self.region, 5 * MB)
        s3_request = s3_client.make_request(
            request=request,
            type=S3RequestType.GET_OBJECT,
            on_headers=self._on_request_headers,
            on_body=_explode,
        )

        finished_future = s3_request.finished_future
        shutdown_event = s3_request.shutdown_event
        s3_request = None
        self.assertTrue(shutdown_event.wait(self.timeout))

        e = finished_future.exception()
        # check that data from on_done callback came through correctly
        self.assertIsInstance(e, RuntimeError)
        self.assertEqual(str(e), "Error in on_body callback")

    def test_special_filepath_upload(self):
        # remove the input file when request done
        content_length = 10 * MB
        special_path = self.files.create_file_with_size(self.special_path, content_length)

        request = self._put_object_request(None, content_length)
        s3_client = s3_client_new(False, self.region, 5 * MB)
        request_type = S3RequestType.PUT_OBJECT

        event_loop_group = EventLoopGroup()
        host_resolver = DefaultHostResolver(event_loop_group)
        bootstrap = ClientBootstrap(event_loop_group, host_resolver)
        credential_provider = AwsCredentialsProvider.new_default_chain(bootstrap)
        # Let signer to normalize uri path for us.
        signing_config = AwsSigningConfig(
            algorithm=AwsSigningAlgorithm.V4,
            signature_type=AwsSignatureType.HTTP_REQUEST_HEADERS,
            service="s3",
            signed_body_header_type=AwsSignedBodyHeaderType.X_AMZ_CONTENT_SHA_256,
            signed_body_value=AwsSignedBodyValue.UNSIGNED_PAYLOAD,
            region=self.region,
            credentials_provider=credential_provider,
            use_double_uri_encode=False,
            should_normalize_uri_path=True,
        )

        s3_request = s3_client.make_request(
            request=request,
            type=request_type,
            send_filepath=special_path,
            signing_config=signing_config,
            on_headers=self._on_request_headers,
            on_progress=self._on_progress,
            on_done=self._on_request_done)
        finished_future = s3_request.finished_future
        finished_future.result(self.timeout)

        # check result
        self.assertEqual(
            content_length,
            self.transferred_len,
            "the transferred length reported does not match body we sent")
        self._validate_successful_response(request_type is S3RequestType.PUT_OBJECT)
        os.remove(special_path)

    def test_non_ascii_filepath_upload(self):
        # remove the input file when request done
        content_length = 10 * MB
        non_ascii_file_path = self.files.create_file_with_size(self.non_ascii_file_name, content_length)
        request = self._put_object_request(None, content_length)
        s3_client = s3_client_new(False, self.region, 5 * MB)
        request_type = S3RequestType.PUT_OBJECT

        s3_request = s3_client.make_request(
            request=request,
            type=request_type,
            send_filepath=non_ascii_file_path,
            on_headers=self._on_request_headers,
            on_progress=self._on_progress,
            on_done=self._on_request_done)
        finished_future = s3_request.finished_future
        finished_future.result(self.timeout)

        # check result
        self.assertEqual(
            content_length,
            self.transferred_len,
            "the transferred length reported does not match body we sent")
        self._validate_successful_response(request_type is S3RequestType.PUT_OBJECT)

    def test_non_ascii_filepath_download(self):
        non_ascii_file_path = self.files.create_file_with_size(self.non_ascii_file_name, 0)
        request = self._get_object_request(self.get_test_object_path)
        request_type = S3RequestType.GET_OBJECT
        s3_client = s3_client_new(False, self.region, 5 * MB)
        s3_request = s3_client.make_request(
            request=request,
            type=request_type,
            recv_filepath=non_ascii_file_path,
            on_headers=self._on_request_headers,
            on_progress=self._on_progress,
            on_done=self._on_request_done)
        finished_future = s3_request.finished_future
        finished_future.result(self.timeout)

        # Result check
        self.data_len = int(HttpHeaders(self.response_headers).get("Content-Length"))
        file_stats = os.stat(non_ascii_file_path)
        file_len = file_stats.st_size
        self.assertEqual(
            file_len,
            self.transferred_len,
            "the length of written file does not match the transferred length reported")
        self.assertEqual(
            self.data_len,
            self.transferred_len,
            "the transferred length reported does not match the content-length header")
        self.assertEqual(self.response_status_code, 200, "status code is not 200")


if __name__ == '__main__':
    unittest.main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

import awscrt.io

print('if the next statement does not explode, the pip install was successful') 
print('Is alpn supported? {}', awscrt.io.is_alpn_available())



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

import argparse
import json
import logging
import os
import sys


DESCRIPTION = """Print 2 tables in GitHub-flavored Markdown that summarize
an execution of CBMC proofs."""


def get_args():
    """Parse arguments for summarize script."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    for arg in [{
            "flags": ["--run-file"],
            "help": "path to the Litani run.json file",
            "required": True,
    }]:
        flags = arg.pop("flags")
        parser.add_argument(*flags, **arg)
    return parser.parse_args()


def _get_max_length_per_column_list(data):
    ret = [len(item) + 1 for item in data[0]]
    for row in data[1:]:
        for idx, item in enumerate(row):
            ret[idx] = max(ret[idx], len(item) + 1)
    return ret


def _get_table_header_separator(max_length_per_column_list):
    line_sep = ""
    for max_length_of_word_in_col in max_length_per_column_list:
        line_sep += "|" + "-" * (max_length_of_word_in_col + 1)
    line_sep += "|\n"
    return line_sep


def _get_entries(max_length_per_column_list, row_data):
    entries = []
    for row in row_data:
        entry = ""
        for idx, word in enumerate(row):
            max_length_of_word_in_col = max_length_per_column_list[idx]
            space_formatted_word = (max_length_of_word_in_col - len(word)) * " "
            entry += "| " + word + space_formatted_word
        entry += "|\n"
        entries.append(entry)
    return entries


def _get_rendered_table(data):
    table = []
    max_length_per_column_list = _get_max_length_per_column_list(data)
    entries = _get_entries(max_length_per_column_list, data)
    for idx, entry in enumerate(entries):
        if idx == 1:
            line_sep = _get_table_header_separator(max_length_per_column_list)
            table.append(line_sep)
        table.append(entry)
    table.append("\n")
    return "".join(table)


def _get_status_and_proof_summaries(run_dict):
    """Parse a dict representing a Litani run and create lists summarizing the
    proof results.

    Parameters
    ----------
    run_dict
        A dictionary representing a Litani run.


    Returns
    -------
    A list of 2 lists.
    The first sub-list maps a status to the number of proofs with that status.
    The second sub-list maps each proof to its status.
    """
    count_statuses = {}
    proofs = [["Proof", "Status"]]
    for proof_pipeline in run_dict["pipelines"]:
        status_pretty_name = proof_pipeline["status"].title().replace("_", " ")
        try:
            count_statuses[status_pretty_name] += 1
        except KeyError:
            count_statuses[status_pretty_name] = 1
        if proof_pipeline["name"] == "print_tool_versions":
            continue
        proofs.append([proof_pipeline["name"], status_pretty_name])
    statuses = [["Status", "Count"]]
    for status, count in count_statuses.items():
        statuses.append([status, str(count)])
    return [statuses, proofs]


def print_proof_results(out_file):
    """
    Print 2 strings that summarize the proof results.
    When printing, each string will render as a GitHub flavored Markdown table.
    """
    output = "## Summary of CBMC proof results\n\n"
    with open(out_file, encoding='utf-8') as run_json:
        run_dict = json.load(run_json)
    status_table, proof_table = _get_status_and_proof_summaries(run_dict)
    for summary in (status_table, proof_table):
        output += _get_rendered_table(summary)

    print(output)
    sys.stdout.flush()

    github_summary_file = os.getenv("GITHUB_STEP_SUMMARY")
    if github_summary_file:
        with open(github_summary_file, "a") as handle:
            print(output, file=handle)
            handle.flush()
    else:
        logging.warning(
            "$GITHUB_STEP_SUMMARY not set, not writing summary file")

    msg = (
        "Click the 'Summary' button to view a Markdown table "
        "summarizing all proof results")
    if run_dict["status"] != "success":
        logging.error("Not all proofs passed.")
        logging.error(msg)
        sys.exit(1)
    logging.info(msg)


if __name__ == '__main__':
    args = get_args()
    logging.basicConfig(format="%(levelname)s: %(message)s")
    try:
        print_proof_results(args.run_file)
    except Exception as ex: # pylint: disable=broad-except
        logging.critical("Could not print results. Exception: %s", str(ex))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import pytest

from configuration import available_ports, TLS13_CIPHERS
from common import ProviderOptions, Protocols, random_str
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N, OpenSSL
from utils import invalid_test_parameters, get_parameter_name

SERVER_DATA = "Some random data from the server:" + random_str(10)
CLIENT_DATA = "Some random data from the client:" + random_str(10)


def test_nothing():
    """
    Sometimes the key update test parameters in combination with the s2n libcrypto
    results in no test cases existing. In this case, pass a nothing test to avoid
    marking the entire codebuild run as failed.
    """
    assert True


@pytest.mark.flaky(reruns=5)
@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL], ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
def test_s2n_server_key_update(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
):
    host = "localhost"
    port = next(available_ports)

    update_requested = b"K"
    starting_marker = "Verify return code"
    key_update_marker = "KEYUPDATE"

    send_marker_list = [starting_marker, key_update_marker]

    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        host=host,
        port=port,
        cipher=cipher,
        data_to_send=[update_requested, CLIENT_DATA.encode()],
        insecure=True,
        protocol=protocol,
    )

    server_options = copy.copy(client_options)

    server_options.mode = Provider.ServerMode
    server_options.key = "../pems/ecdsa_p384_pkcs1_key.pem"
    server_options.cert = "../pems/ecdsa_p384_pkcs1_cert.pem"
    server_options.data_to_send = [SERVER_DATA.encode()]

    server = managed_process(S2N, server_options, send_marker=CLIENT_DATA, timeout=30)
    client = managed_process(
        provider,
        client_options,
        send_marker=send_marker_list,
        close_marker=SERVER_DATA,
        timeout=30,
    )

    for results in client.get_results():
        results.assert_success()
        assert key_update_marker in str(results.stderr)
        assert SERVER_DATA.encode() in results.stdout

    for results in server.get_results():
        results.assert_success()
        assert CLIENT_DATA.encode() in results.stdout


@pytest.mark.flaky(reruns=5)
@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL], ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
def test_s2n_client_key_update(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
):
    host = "localhost"
    port = next(available_ports)

    update_requested = b"K\n"
    # Last statement printed out by Openssl after handshake
    starting_marker = "Secure Renegotiation IS supported"
    key_update_marker = "TLSv1.3 write server key update"
    read_key_update_marker = b"TLSv1.3 read client key update"

    send_marker_list = [starting_marker, key_update_marker]

    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        host=host,
        port=port,
        cipher=cipher,
        data_to_send=[CLIENT_DATA.encode()],
        insecure=True,
        protocol=protocol,
    )

    server_options = copy.copy(client_options)

    server_options.mode = Provider.ServerMode
    server_options.key = "../pems/ecdsa_p384_pkcs1_key.pem"
    server_options.cert = "../pems/ecdsa_p384_pkcs1_cert.pem"
    server_options.data_to_send = [update_requested, SERVER_DATA.encode()]

    server = managed_process(
        provider,
        server_options,
        send_marker=send_marker_list,
        close_marker=CLIENT_DATA,
        timeout=30,
    )
    client = managed_process(
        S2N,
        client_options,
        send_marker=SERVER_DATA,
        close_marker=SERVER_DATA,
        timeout=30,
    )

    for results in client.get_results():
        results.assert_success()
        assert SERVER_DATA.encode() in results.stdout

    for results in server.get_results():
        results.assert_success()
        assert read_key_update_marker in results.stderr
        assert CLIENT_DATA.encode() in results.stdout



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import os

from . import github
from . import sns
from datetime import datetime, timedelta
from dateutil import parser, tz

logger = logging.getLogger()
logger.setLevel(logging.INFO)
# What time range to consider alerting on failures.
TIME_WINDOW_BEGIN = datetime.now().astimezone(tz.UTC) - timedelta(hours=float(os.getenv('MONITOR_FREQ_IN_HOURS')))
TIME_WINDOW_END = datetime.now().astimezone(tz.UTC)


class GitHubActions(github.GitHubClient):
    params = {
        # Needed when using an API key
        'github_username': os.getenv('github_username', None),
        'github_password': os.getenv('github_password', None),
        # Use from within an Action - ignored if username set
        'token': os.getenv('GITHUB_TOKEN', None),
        'repo_organization': os.getenv('GITHUB_REPO_ORG'),
        'repo': os.getenv('GITHUB_REPO')
    }


class S2nNotices(sns.SNSClient):
    params = {
        'topic_arn': 'arn:aws:sns:us-west-2:024603541914:s2n_notices'
    }


def message_text():
    """ Formatting for text message. """
    return """
s2n GitHub Action monitor notice
State: {conclusion}
Repo: {repo}
GHA failure time: {time}
Workflow name: {workflow_name}
URL: {url}
started by: {commit_owner}\n
"""


def main():
    """ Main entrypoint. """
    logging.info('Starting up')
    plaintext_notice = []
    gh_api = GitHubActions()
    s2n_text_client = S2nNotices()

    # Get the Action workflow log from the Github API
    gh_api.get_workflow_log_chunk(final_state='failure')
    logging.info(f"Looking for failures newer than {TIME_WINDOW_BEGIN} in {gh_api.params['repo_organization']}"
                 f"/{gh_api.params['repo']}")
    if gh_api.worklog:
        for enhanced_worklog in gh_api.worklog:
            # Parse the event date/time so we can compare it
            datetime_creation = parser.parse(enhanced_worklog['created_at'])
            logging.debug(f"looking at event from {enhanced_worklog['created_at']}")

            # If the event is recent enough, process it.
            if datetime_creation > TIME_WINDOW_BEGIN:
                logging.debug(f"Workflow_url: {enhanced_worklog['workflow_url']}")
                # The name of the workflow isn't in the failure object, look it up.
                enhanced_worklog['workflow_name'] = gh_api.get_workflow_name(enhanced_worklog['workflow_url'].split('/')[-1:][0])
                enhanced_worklog['repo'] = gh_api.params['repo']

                # Construct a notification string.
                notice_msg = message_text().format(
                    conclusion=enhanced_worklog['conclusion'],
                    time=enhanced_worklog['created_at'],
                    url=enhanced_worklog['html_url'],
                    commit_owner=enhanced_worklog['head_commit']['author']['email'],
                    repo=enhanced_worklog['repo'],
                    workflow_name=enhanced_worklog['workflow_name'])
                logging.debug(notice_msg)
                plaintext_notice.append(notice_msg)
            else:
                logging.debug("event outside time range.")
    else:
        logging.info("GH API returned empty worklog")


    # Relay messages to SNS
    if plaintext_notice:
        # Combine multiple message together.
        logging.info(s2n_text_client.publish("\n".join(plaintext_notice)))
        logging.info(f"Notices published")
    logging.info("Done")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Minor change based on the example from hyper h2 server
# https://github.com/python-hyper/h2/blob/master/examples/asyncio/asyncio-server.py

"""
asyncio-server.py
~~~~~~~~~~~~~~~~~

A fully-functional HTTP/2 server using asyncio. Requires Python 3.5+.

This example demonstrates handling requests with bodies, as well as handling
those without. In particular, it demonstrates the fact that DataReceived may
be called multiple times, and that applications must handle that possibility.
"""
import asyncio
import io
import json
import ssl
import time
import os
import collections
from typing import List, Tuple

from h2.config import H2Configuration
from h2.connection import H2Connection
from h2.events import (
    ConnectionTerminated, DataReceived, RemoteSettingsChanged,
    RequestReceived, StreamEnded, StreamReset, WindowUpdated
)
from h2.errors import ErrorCodes
from h2.exceptions import ProtocolError, StreamClosedError
from h2.settings import SettingCodes


RequestData = collections.namedtuple('RequestData', ['headers', 'data'])


class H2Protocol(asyncio.Protocol):
    def __init__(self):
        config = H2Configuration(client_side=False, header_encoding='utf-8')
        self.conn = H2Connection(config=config)
        self.transport = None
        self.stream_data = {}
        self.flow_control_futures = {}
        self.file_path = None
        self.num_sentence_received = {}
        self.raw_headers = None
        self.download_test_length = 2500000000
        self.out_bytes_per_second = 900

    def connection_made(self, transport: asyncio.Transport):
        self.transport = transport
        self.conn.initiate_connection()
        # Increase the window to large enough to avoid blocking on the window limits.
        self.conn.increment_flow_control_window(int(2147483647/2))
        self.transport.write(self.conn.data_to_send())

    def connection_lost(self, exc):
        for future in self.flow_control_futures.values():
            future.cancel()
        self.flow_control_futures = {}

    def data_received(self, data: bytes):
        try:
            events = self.conn.receive_data(data)
        except ProtocolError as e:
            self.transport.write(self.conn.data_to_send())
            self.transport.close()
        else:
            self.transport.write(self.conn.data_to_send())
            for event in events:
                if isinstance(event, RequestReceived):
                    self.request_received(event.headers, event.stream_id)
                elif isinstance(event, DataReceived):
                    self.receive_data(event.data, event.stream_id,
                                      event.flow_controlled_length)
                elif isinstance(event, StreamEnded):
                    self.stream_complete(event.stream_id)
                elif isinstance(event, ConnectionTerminated):
                    self.transport.close()
                elif isinstance(event, StreamReset):
                    self.stream_reset(event.stream_id)
                elif isinstance(event, WindowUpdated):
                    self.window_updated(event.stream_id, event.delta)
                elif isinstance(event, RemoteSettingsChanged):
                    if SettingCodes.INITIAL_WINDOW_SIZE in event.changed_settings:
                        self.window_updated(None, 0)

                self.transport.write(self.conn.data_to_send())

    def request_received(self, headers: List[Tuple[str, str]], stream_id: int):
        # Bump the flow control window to large enough to avoid blocking on the window limits.
        self.conn.increment_flow_control_window(
            int(2147483647/2), stream_id)
        self.raw_headers = headers
        headers = collections.OrderedDict(headers)
        path = headers[':path']
        method = headers[':method']
        if method == "PUT" or method == "POST":
            self.file_path = os.path.join(os.path.curdir, path[1:])
            if os.path.exists(self.file_path):
                os.remove(self.file_path)

        # Store off the request data.
        request_data = RequestData(headers, io.BytesIO())
        self.stream_data[stream_id] = request_data

    def handle_request_echo(self, stream_id: int, request_data: RequestData):
        response_headers = [(':status', '200')]
        for i in self.raw_headers:
            # Response headers back and exclude pseudo headers
            if i[0][0] != ':':
                response_headers.append(i)
        body = request_data.data.getvalue().decode('utf-8')
        data = json.dumps(
            {"body": body}, indent=4
        ).encode("utf8")
        self.conn.send_headers(stream_id, response_headers)
        asyncio.ensure_future(self.send_data(data, stream_id))

    def stream_complete(self, stream_id: int):
        """
        When a stream is complete, we can send our response.
        """
        try:
            request_data = self.stream_data[stream_id]
        except KeyError:
            # Just return, we probably 405'd this already
            return

        path = request_data.headers[':path']
        method = request_data.headers[':method']
        if method == "PUT" or method == "POST":
            self.conn.send_headers(stream_id, [(':status', '200')])
            asyncio.ensure_future(self.send_data(
                str(self.num_sentence_received[stream_id]).encode(), stream_id))
        elif path == '/echo':
            self.handle_request_echo(stream_id, request_data)
        elif path == '/downloadTest':
            length = self.download_test_length
            self.conn.send_headers(
                stream_id, [(':status', '200'), ('content-length', str(length))])
            asyncio.ensure_future(self.send_repeat_data(length, stream_id))
        elif path == '/slowConnTest':
            length = int(self.download_test_length/1000)
            self.conn.send_headers(
                stream_id, [(':status', '200'), ('content-length', str(length))])
            asyncio.ensure_future(
                self.send_slow_repeat_data(length, stream_id))
        else:
            self.conn.send_headers(stream_id, [(':status', '404')])
            asyncio.ensure_future(self.send_data(b"Not Found", stream_id))

    def receive_data(self, data: bytes, stream_id: int, flow_controlled_length: int):
        """
        We've received some data on a stream. If that stream is one we're
        expecting data on, save it off. Otherwise, reset the stream.
        """
        try:
            if flow_controlled_length > 0:
                # We need to update the flow control window for the stream
                # and the connection. And the function only accepts value > 0.
                self.conn.increment_flow_control_window(
                    flow_controlled_length)
                self.conn.increment_flow_control_window(
                    flow_controlled_length, stream_id)
            stream_data = self.stream_data[stream_id]
        except KeyError:
            self.conn.reset_stream(
                stream_id, error_code=ErrorCodes.PROTOCOL_ERROR
            )
        else:
            method = stream_data.headers[':method']
            if method == "PUT" or method == "POST":
                if stream_id in self.num_sentence_received:
                    self.num_sentence_received[stream_id] = self.num_sentence_received[stream_id] + \
                        len(data)
                else:
                    self.num_sentence_received[stream_id] = len(data)

            else:
                stream_data.data.write(data)

    def stream_reset(self, stream_id):
        """
        A stream reset was sent. Stop sending data.
        """
        if stream_id in self.flow_control_futures:
            future = self.flow_control_futures.pop(stream_id)
            future.cancel()

    async def send_data(self, data, stream_id):
        """
        Send data according to the flow control rules.
        """
        while data:
            while self.conn.local_flow_control_window(stream_id) < 1:
                try:
                    await self.wait_for_flow_control(stream_id)
                except asyncio.CancelledError:
                    return

            chunk_size = min(
                self.conn.local_flow_control_window(stream_id),
                len(data),
                self.conn.max_outbound_frame_size,
            )

            try:
                self.conn.send_data(
                    stream_id,
                    data[:chunk_size],
                    end_stream=(chunk_size == len(data))
                )
            except (StreamClosedError, ProtocolError):
                # The stream got closed and we didn't get told. We're done
                # here.
                break

            self.transport.write(self.conn.data_to_send())
            data = data[chunk_size:]

    async def send_repeat_data(self, length, stream_id):
        """
        Send data with length according to the flow control rules.
        """
        while length > 0:
            while self.conn.local_flow_control_window(stream_id) < 1:
                try:
                    await self.wait_for_flow_control(stream_id)
                except asyncio.CancelledError:
                    return

            chunk_size = min(
                self.conn.local_flow_control_window(stream_id),
                length,
                self.conn.max_outbound_frame_size,
            )
            repeated = b"This is CRT HTTP test."
            data = int(chunk_size/len(repeated)) * repeated + \
                repeated[:chunk_size % len(repeated)]

            try:
                self.conn.send_data(
                    stream_id,
                    data,
                    end_stream=(chunk_size == length)
                )
            except (StreamClosedError, ProtocolError):
                # The stream got closed and we didn't get told. We're done
                # here.
                break

            self.transport.write(self.conn.data_to_send())
            length = length - chunk_size

    async def send_slow_repeat_data(self, length, stream_id):
        """
        Send data with length slowly (less than 1000 bytes per second)
        """
        while length > 0:
            while self.conn.local_flow_control_window(stream_id) < 1:
                try:
                    await self.wait_for_flow_control(stream_id)
                except asyncio.CancelledError:
                    return

            chunk_size = min(
                self.conn.local_flow_control_window(stream_id),
                length,
                self.conn.max_outbound_frame_size,
                self.out_bytes_per_second
            )
            repeated = b"This is CRT HTTP test."
            data = int(chunk_size/len(repeated)) * repeated + \
                repeated[:chunk_size % len(repeated)]

            try:
                # Sleep for a sec to make the out bytes per second slower than the expected
                time.sleep(1)
                self.conn.send_data(
                    stream_id,
                    data,
                    end_stream=(chunk_size == length)
                )
            except (StreamClosedError, ProtocolError):
                # The stream got closed and we didn't get told. We're done
                # here.
                break

            self.transport.write(self.conn.data_to_send())
            length = length - chunk_size

    async def wait_for_flow_control(self, stream_id):
        """
        Waits for a Future that fires when the flow control window is opened.
        """
        f = asyncio.Future()
        self.flow_control_futures[stream_id] = f
        await f

    def window_updated(self, stream_id, delta):
        """
        A window update frame was received. Unblock some number of flow control
        Futures.
        """
        if stream_id and stream_id in self.flow_control_futures:
            f = self.flow_control_futures.pop(stream_id)
            f.set_result(delta)
        elif not stream_id:
            for f in self.flow_control_futures.values():
                f.set_result(delta)

            self.flow_control_futures = {}


ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
ssl_context.options |= (ssl.OP_NO_COMPRESSION)
ssl_context.load_cert_chain(
    certfile=os.path.join(os.path.dirname(__file__),
                          "../resources/unittests.crt"),
    keyfile=os.path.join(os.path.dirname(__file__), "../resources/unittests.key"))
ssl_context.set_alpn_protocols(["h2"])

loop = asyncio.new_event_loop()
# Each client connection will create a new protocol instance
coro = loop.create_server(H2Protocol, '127.0.0.1', 3443, ssl=ssl_context)
server = loop.run_until_complete(coro)

# Serve requests until Ctrl+C is pressed
print('Serving on {}'.format(server.sockets[0].getsockname()))
try:
    loop.run_forever()
except KeyboardInterrupt:
    pass

# Close the server
server.close()
loop.run_until_complete(server.wait_closed())
loop.close()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.
import argparse
import sys
import os
from io import BytesIO
from awscrt import io, http
from urllib.parse import urlparse


def print_header_list(headers):
    for name, value in headers:
        print('{}: {}'.format(name, value))


parser = argparse.ArgumentParser()
parser.add_argument(
    'url',
    help='URL to make request to. HTTPS is assumed unless port 80 is specified or HTTP is specified in the scheme.')
parser.add_argument('--cacert', required=False, help='FILE: path to a CA certificate file.')
parser.add_argument('--capath', required=False, help='PATH: path to a directory containing CA files.')
parser.add_argument('--cert', required=False, help='FILE: path to a PEM encoded certificate to use with mTLS')
parser.add_argument('--key', required=False, help='FILE: Path to a PEM encoded private key that matches cert.')
parser.add_argument(
    '--connect_timeout',
    required=False,
    type=int,
    help='INT: time in milliseconds to wait for a connection.',
    default=5000)
parser.add_argument(
    '-H',
    '--header',
    required=False,
    help='STRING: line to send as a header in format "name:value". May be specified multiple times.',
    action='append')
parser.add_argument('-d', '--data', required=False, help='STRING: Data to POST or PUT.')
parser.add_argument('--data_file', required=False, help='FILE: File to read from file and POST or PUT')
parser.add_argument(
    '-M',
    '--method',
    required=False,
    help='STRING: Http Method verb to use for the request',
    default='GET')
parser.add_argument('-G', '--get', required=False, help='uses GET for the verb', action='store_true')
parser.add_argument('-P', '--post', required=False, help='uses POST for the verb', action='store_true')
parser.add_argument('-I', '--head', required=False, help='uses HEAD for the verb', action='store_true')
parser.add_argument(
    '-i',
    '--include',
    required=False,
    help='Includes headers in output',
    action='store_true',
    default=False)
parser.add_argument(
    '-k',
    '--insecure',
    required=False,
    help='Turns off x.509 validation',
    action='store_true',
    default=False)
parser.add_argument('-o', '--output', required=False, help='FILE: dumps content-body to FILE instead of stdout.')
parser.add_argument('-t', '--trace', required=False, help='FILE: dumps logs to FILE instead of stderr.')
parser.add_argument(
    '-p',
    '--alpn',
    default=["h2", "http/1.1"],
    required=False,
    help='STRING: protocol for ALPN. May be specified multiple times.',
    action='append')
parser.add_argument(
    '--http2',
    required=False,
    help='HTTP/2 connection required',
    action="store_true")
parser.add_argument(
    '--http1_1',
    required=False,
    help='HTTP/1.1 connection required',
    action="store_true")
parser.add_argument(
    '-v',
    '--verbose',
    required=False,
    help='ERROR|INFO|DEBUG|TRACE: log level to configure. Default is none.')

args = parser.parse_args()

output = getattr(sys.stdout, 'buffer', sys.stdout)

if args.output:
    output = open(args.output, mode='wb')

# setup the logger if user request logging

if args.verbose:
    log_level = io.LogLevel.NoLogs

    if args.verbose == 'ERROR':
        log_level = io.LogLevel.Error
    elif args.verbose == 'INFO':
        log_level = io.LogLevel.Info
    elif args.verbose == 'DEBUG':
        log_level = io.LogLevel.Debug
    elif args.verbose == 'TRACE':
        log_level = io.LogLevel.Trace
    else:
        print('{} unsupported value for the verbose option'.format(args.verbose))
        exit(-1)

    log_output = 'stderr'

    if args.trace:
        log_output = args.trace

    io.init_logging(log_level, log_output)


required_version = http.HttpVersion.Unknown
if args.http1_1:
    required_version = http.HttpVersion.Http1_1
    args.alpn = ["http/1.1"]
if args.http2:
    required_version = http.HttpVersion.Http2
    args.alpn = ["h2"]


# an event loop group is needed for IO operations. Unless you're a server or a client doing hundreds of connections
# you only want one of these.
event_loop_group = io.EventLoopGroup(1)

host_resolver = io.DefaultHostResolver(event_loop_group)

# client bootstrap knows how to connect all the pieces. In this case it also has the default dns resolver
# baked in.
client_bootstrap = io.ClientBootstrap(event_loop_group, host_resolver)

url = urlparse(args.url)
port = 443
scheme = 'https'

if url.scheme is not None and url.scheme == 'http':
    scheme = 'http'

if url.port is not None:
    port = url.port
else:
    if scheme == 'http':
        port = 80
        if args.http2:
            sys.exit("Error, we don't support h2c, please use TLS for HTTP/2 connection")


tls_connection_options = None

if scheme == 'https':
    if args.cert is not None and args.key is not None:
        tls_ctx_options = io.TlsContextOptions.create_client_with_mtls_from_path(args.cert, args.key)
    else:
        tls_ctx_options = io.TlsContextOptions()

    if args.cacert is not None or args.capath is not None:
        tls_ctx_options.override_default_trust_store_from_path(args.capath, args.cacert)

    if args.insecure:
        tls_ctx_options.verify_peer = False

    tls_ctx = io.ClientTlsContext(tls_ctx_options)

    tls_connection_options = tls_ctx.new_connection_options()
    tls_connection_options.set_server_name(url.hostname)

    if args.alpn:
        tls_connection_options.set_alpn_list(args.alpn)

# invoked up on the connection closing


def on_connection_shutdown(shutdown_future):
    print('connection close with error: {}'.format(shutdown_future.exception()))


# invoked by the http request call as the response body is received in chunks
def on_incoming_body(http_stream, chunk, **kwargs):
    output.write(chunk)


data_len = 0
data_stream = None

if args.data:
    data_bytes = args.data.encode(encoding='utf-8')
    data_len = len(data_bytes)
    data_stream = BytesIO(data_bytes)
elif args.data_file:
    data_len = os.stat(args.data_file).st_size
    data_stream = open(args.data_file, 'rb')


socket_options = io.SocketOptions()
socket_options.connect_timeout_ms = args.connect_timeout

hostname = url.hostname
connect_future = http.HttpClientConnection.new(
    host_name=hostname,
    port=port,
    socket_options=socket_options,
    tls_connection_options=tls_connection_options,
    bootstrap=client_bootstrap)

connection = connect_future.result(10)
connection.shutdown_future.add_done_callback(on_connection_shutdown)

if required_version:
    if connection.version != required_version:
        error_msg = "Error. The requested HTTP version " + args.alpn[0] + " is not supported by the peer."
        sys.exit(error_msg)


request = http.HttpRequest(args.method, body_stream=data_stream)

if args.get:
    request.method = "GET"

if args.post:
    request.method = "POST"

if args.head:
    request.method = "HEAD"

if url.path:
    request.path = url.path

if url.query:
    request.path += '?' + url.query

if connection.version == http.HttpVersion.Http2:
    request.headers.add(':authority', hostname)
else:
    request.headers.add('host', hostname)
request.headers.add('user-agent', 'elasticurl.py 1.0, Powered by the AWS Common Runtime.')

if data_len != 0:
    request.headers.add('content-length', str(data_len))

if args.header:
    for i in args.header:
        name, value = i.split(':')
        request.headers.add(name.strip(), value.strip())

# invoked as soon as the response headers are received


def response_received_cb(http_stream, status_code, headers, **kwargs):
    if args.include:
        print('Response Code: {}'.format(status_code))
        print_header_list(headers)


# make the request
stream = connection.request(request, response_received_cb, on_incoming_body)
stream.activate()

# wait until the full response is finished
stream.completion_future.result()
stream = None
connection = None

if data_stream:
    data_stream.close()

if args.output:
    output.close()



"""
Setup local TLS server for tests
"""

import Builder

import os
import sys
import subprocess
import atexit
import time


class TlsServerSetup(Builder.Action):
    """
    Set up this machine for running the mock server test

    This action should be run in the 'pre_build_steps' or 'build_steps' stage.
    """

    def run(self, env):
        if not env.project.needs_tests(env):
            print("Skipping TLS server setup because tests disabled for project")
            return

        self.env = env

        base_dir = os.path.dirname(os.path.realpath(__file__))
        dir = os.path.join(base_dir, "..", "..", "tests", "tls_server")

        print("Running openssl TLS server")

        python_path = sys.executable
        p = subprocess.Popen([python_path, "tls_server.py",
                              ], cwd=dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        @atexit.register
        def close_tls_server():
            print("Terminating openssl TLS server")
            p.terminate()
            out, err = p.communicate()
            print("TLS server stdout:\n{}".format(out))
            print("TLS server stderr:\n{}".format(err))



# SCRAM mode Python script
import sys
import hmac
import hashlib
from Crypto.Cipher import AES
from Crypto import Random
rndfile = Random.new()

DEBUG_ENABLED = True

# When reading/writing byte strings, the first (aka left-most) byte is the Most Significant Byte (aka Big-Endian)
# (Eg "0x0001", 0x00 is the MSB and 0x01 is the LSB, meaning 0x0001 == 1)
ENDIANNESS = 'big'


def byteStr(val, numBytes):
    return val.to_bytes(numBytes, ENDIANNESS)


def debugByteStr(debugStr, byteStrVal):
    if DEBUG_ENABLED:
        print(debugStr + ": 0x" + byteStrVal.hex().upper())


def debugInt(debugStr, intVal):
    if DEBUG_ENABLED:
        print(debugStr + ": " + str(intVal))


def scram_generate_key():
    # Generate Random 32 Byte Key
    K = rndfile.read(32)
    debugByteStr("K", K)

    return K


def scram_encrypt(K, N, A, M, F):
    """
    SCRAM Encryption

    Parameters:
            K: Key
            N: Nonce
            A: Additional Authenticated Data
            M: Plaintext Message
            F: Frame Size

    Returns:
            C: Ciphertext
            X: Excrypted R and Padding Len
            Tag: Authentication Tag
    """
    # Generate a random 32-byte value R
    R = rndfile.read(32)

    # Prepare the Padding. We append 0x00 bytes to the end up to the next frame size.
    M_LEN = len(M)
    PADDING_LEN = 0

    if (F > 0):
        PADDING_LEN = (F - M_LEN) % F

    PADDING_STR = byteStr(0x0, PADDING_LEN)
    PADDING_LEN_STR = byteStr(PADDING_LEN, 2)
    PADDED_MSG = M + PADDING_STR

    debugInt("len(M)", M_LEN)
    debugInt("PADDING_LEN", PADDING_LEN)
    debugByteStr("PADDING_STR", PADDING_STR)
    debugByteStr("PADDING_LEN_STR", PADDING_LEN_STR)
    debugByteStr("PADDED_MSG", PADDED_MSG)

    # Derive Message encryption key (KE)
    # S1 = N || 0x00 0x00 0x00 0x1 || 0^{8} || 0^{8} || 0^{16} || R
    S1 = N + byteStr(0x01, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + byteStr(0x0, 16) + R
    U1 = hmac.new(K, S1, hashlib.sha512).digest()
    KE = U1[0:32]

    # AES_CTR encrypt PADDED_MSG with Nonce N and Key KE
    C = AES.new(key=KE, mode=AES.MODE_CTR, nonce=N).encrypt(PADDED_MSG)

    # Derive MAC Key (KM) used to with GMAC to generate T
    # S2 = N || 0x00 0x00 0x00 0x2 || 0^{8} || 0^{8} || 0^{16} || 0^{32}
    S2 = N + byteStr(0x02, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + byteStr(0x0, 16) + byteStr(0x0, 32)
    U2 = hmac.new(K, S2, hashlib.sha512).digest()
    KM = U2[0:32]

    # GMAC the string A || C , using the GMAC key KM and nonce N
    T = AES.new(key=KM, mode=AES.MODE_GCM, nonce=N).update(A + C).digest()

    # Derive a one-time pad (U3) from T
    # S3 = N || 0x00 0x00 0x00 0x3 || 0^{8} || 0^{8} || T || 0^{32}
    S3 = N + byteStr(0x03, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + T + byteStr(0x0, 32)
    U3 = hmac.new(K, S3, hashlib.sha512).digest()

    # Encrypt R and PaddingLen with one-time pad U3
    Y1 = bytes(a ^ b for (a, b) in zip(U3[0:32], R))
    Y0 = bytes(a ^ b for (a, b) in zip(U3[32:34], PADDING_LEN_STR))
    X = Y1 + Y0

    # Authenticate (Tag) T and R
    # S4 = N || 0x00 0x00 0x00 0x4 || A_LEN_STR || M_LEN_STR || T || R
    S4 = N + byteStr(0x04, 4) + byteStr(len(A), 8) + byteStr(M_LEN, 8) + T + R
    U4 = hmac.new(K, S4, hashlib.sha512).digest()

    # Truncate to 16 bytes tag
    Tag = U4[0:16]

    debugByteStr("S1", S1)
    debugByteStr("S2", S2)
    debugByteStr("S3", S3)
    debugByteStr("S4", S4)
    debugByteStr("U1", U1)
    debugByteStr("U2", U2)
    debugByteStr("U3", U3)
    debugByteStr("U4", U4)
    debugByteStr("Y0", Y0)
    debugByteStr("Y1", Y1)
    debugByteStr("T", T)
    debugByteStr("KE", KE)
    debugByteStr("KM", KM)
    debugInt("len(C)", len(C))
    debugByteStr("C", C)
    debugByteStr("X", X)
    debugByteStr("Tag", Tag)

    return C, X, Tag


def scram_decrypt(K, N, A, C, X, Tag):
    """
    SCRAM Decryption

    Parameters:
            K: Key
            N: Nonce
            A: Additional Authenticated Data
            C: Ciphertext
            X: Encrypted Random value R and Padding Length
            Tag: Tag

    Returns:
            M_calculated: The decrypted Message
    """

    # Derive MAC key (KM)
    # S2 = N || 0x00 0x00 0x00 0x2 || 0^{8} || 0^{8} || 0^{16} || 0^{32}
    S2_calculated = N + byteStr(0x02, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + byteStr(0x0, 16) + byteStr(0x0, 32)
    U2_calculated = hmac.new(K, S2_calculated, hashlib.sha512).digest()
    KM_calculated = U2_calculated[0:32]

    # Derive T
    # T = GMAC (N, A||C, null)
    T_calculated = AES.new(key=KM_calculated, mode=AES.MODE_GCM, nonce=N).update(A + C).digest()

    # Derive one-time pad U3 from T_calculated,
    # S3 = N || 0x00 0x00 0x00 0x3 || 0^{8} || 0^{8} || T || 0^{32}
    S3_calculated = N + byteStr(0x03, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + T_calculated + byteStr(0x0, 32)
    U3_calculated = hmac.new(K, S3_calculated, hashlib.sha512).digest()

    # Decrypt R and PADDING_LEN, by xor'ing X and U3
    R_calculated = bytes(a ^ b for (a, b) in zip(U3_calculated[0:32], X[0:32]))
    PADDING_LEN_STR_calculated = bytes(a ^ b for (a, b) in zip(U3_calculated[32:34], X[32:34]))

    # Derive Message and Padding Lengths
    PADDING_LEN_calculated = int.from_bytes(PADDING_LEN_STR_calculated, ENDIANNESS)
    M_LEN_calculated = len(C) - PADDING_LEN_calculated

    # Authenticate R
    # S4 = N || 0x00 0x00 0x00 0x4 || A_LEN_STR || M_LEN_STR || T || R
    S4_calculated = N + byteStr(0x04, 4) + byteStr(len(A), 8) + \
        byteStr(M_LEN_calculated, 8) + T_calculated + R_calculated
    U4_calculated = hmac.new(K, S4_calculated, hashlib.sha512).digest()
    Tag_calculated = U4_calculated[0:16]

    if (Tag == Tag_calculated):
        print("PASSED: Authentication")
    else:
        print("FAILED: Authentication")
        return None

    # Now that Ciphertext and other parameters are authenticated, we can decrypt Ciphertext to get Plaintext
    # Derive Message Encryption key (KE)
    # S1 = N || 0x00 0x00 0x00 0x1 || 0^{8} || 0^{8} || 0^{16} || R
    S1_calculated = N + byteStr(0x01, 4) + byteStr(0x0, 8) + byteStr(0x0, 8) + byteStr(0x0, 16) + R_calculated
    U1_calculated = hmac.new(K, S1_calculated, hashlib.sha512).digest()
    KE_calculated = U1_calculated[0:32]

    # Decrypt Ciphertext
    PADDED_MSG_calculated = AES.new(key=KE_calculated, mode=AES.MODE_CTR, nonce=N).decrypt(C)

    # Strip off padding bytes
    M_calculated = PADDED_MSG_calculated[0:M_LEN_calculated]

    if DEBUG_ENABLED:
        print("\nDecryption Debug Info: ")
        debugByteStr("S1_calculated", S1_calculated)
        debugByteStr("S2_calculated", S2_calculated)
        debugByteStr("S3_calculated", S3_calculated)
        debugByteStr("S4_calculated", S4_calculated)
        debugByteStr("U1_calculated", U1_calculated)
        debugByteStr("U2_calculated", U2_calculated)
        debugByteStr("U3_calculated", U3_calculated)
        debugByteStr("U4_calculated", U4_calculated)
        debugByteStr("T_calculated", T_calculated)
        debugByteStr("R_calculated", R_calculated)
        debugByteStr("KE_calculated", KE_calculated)
        debugByteStr("KM_calculated", KM_calculated)
        debugByteStr("PADDED_MSG_calculated", PADDED_MSG_calculated)
        debugByteStr("M_calculated", M_calculated)

    return M_calculated


def main(argv):
    # Generate Random 28 Byte Message
    M = rndfile.read(28)
    debugByteStr("M", M)

    # Generate Random 28 Byte Additional Authenticated Data
    A = rndfile.read(28)
    debugByteStr("A", A)

    # Generate Random 12 Byte Key
    N = rndfile.read(12)

    # Frame Size. Messages will be padded up to the next Frame size before being encrypted.
    F = 32
    debugInt("F", F)

    K = scram_generate_key()

    C, X, Tag = scram_encrypt(K, N, A, M, F)

    M_calculated = scram_decrypt(K, N, A, C, X, Tag)

    if (M != M_calculated):
        print("FAILED: Decryption")
    else:
        print("PASSED: Decryption")

    return


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
TEST_CERT_DIRECTORY = "../pems/"
TEST_SNI_CERT_DIRECTORY = "../pems/sni/"
TEST_OCSP_DIRECTORY = "../pems/ocsp/"

TRUST_STORE_BUNDLE = TEST_CERT_DIRECTORY + "trust-store/ca-bundle.crt"
TRUST_STORE_TRUSTED_BUNDLE = TEST_CERT_DIRECTORY + "trust-store/ca-bundle.trust.crt"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import collections

from common import Certificates, Ciphers, Curves, Protocols, AvailablePorts
from constants import TEST_SNI_CERT_DIRECTORY
from providers import S2N, OpenSSL, JavaSSL


# The boolean configuration will let a test run for True and False
# for some value. For example, using the insecure flag.
BOOLEAN = [True, False]


# List of all protocols to be tested
PROTOCOLS = [
    Protocols.TLS13,
    Protocols.TLS12,
    Protocols.TLS11,
    Protocols.TLS10,
    Protocols.SSLv3,
]


# List of providers that will be tested.
PROVIDERS = [S2N, OpenSSL, JavaSSL]


# List of binary TLS13 settings
TLS13 = [True, False]


# List of all curves that will be tested.
ALL_TEST_CURVES = [Curves.X25519, Curves.P256, Curves.P384, Curves.P521]


# List of all certificates that will be tested.
ALL_TEST_CERTS = [
    Certificates.RSA_1024_SHA256,
    Certificates.RSA_1024_SHA384,
    Certificates.RSA_1024_SHA512,
    Certificates.RSA_2048_SHA256,
    Certificates.RSA_2048_SHA384,
    Certificates.RSA_2048_SHA512,
    Certificates.RSA_3072_SHA256,
    Certificates.RSA_3072_SHA384,
    Certificates.RSA_3072_SHA512,
    Certificates.RSA_4096_SHA256,
    Certificates.RSA_4096_SHA384,
    Certificates.RSA_4096_SHA512,
    Certificates.ECDSA_256,
    Certificates.ECDSA_384,
    Certificates.ECDSA_521,
    Certificates.RSA_PSS_2048_SHA256,
]


# List of certificates which enable all cipher suites
MINIMAL_TEST_CERTS = [
    Certificates.RSA_2048_SHA512,
    Certificates.RSA_4096_SHA256,
    Certificates.ECDSA_256,
    Certificates.ECDSA_384,
    Certificates.RSA_PSS_2048_SHA256,
    Certificates.ECDSA_521,
]


# List of all ciphers that will be tested.
ALL_TEST_CIPHERS = [
    Ciphers.DHE_RSA_AES128_SHA,
    Ciphers.DHE_RSA_AES256_SHA,
    Ciphers.DHE_RSA_AES128_SHA256,
    Ciphers.DHE_RSA_AES256_SHA256,
    Ciphers.DHE_RSA_AES128_GCM_SHA256,
    Ciphers.DHE_RSA_AES256_GCM_SHA384,
    Ciphers.DHE_RSA_CHACHA20_POLY1305,
    Ciphers.AES128_SHA,
    Ciphers.AES256_SHA,
    Ciphers.AES128_SHA256,
    Ciphers.AES256_SHA256,
    Ciphers.AES128_GCM_SHA256,
    Ciphers.AES256_GCM_SHA384,
    Ciphers.ECDHE_ECDSA_AES128_GCM_SHA256,
    Ciphers.ECDHE_ECDSA_AES256_GCM_SHA384,
    Ciphers.ECDHE_ECDSA_AES128_SHA256,
    Ciphers.ECDHE_ECDSA_AES256_SHA384,
    Ciphers.ECDHE_ECDSA_AES128_SHA,
    Ciphers.ECDHE_ECDSA_AES256_SHA,
    Ciphers.ECDHE_ECDSA_CHACHA20_POLY1305,
    Ciphers.ECDHE_RSA_AES128_SHA,
    Ciphers.ECDHE_RSA_AES256_SHA,
    Ciphers.ECDHE_RSA_AES128_SHA256,
    Ciphers.ECDHE_RSA_AES256_SHA384,
    Ciphers.ECDHE_RSA_AES128_GCM_SHA256,
    Ciphers.ECDHE_RSA_AES256_GCM_SHA384,
    Ciphers.ECDHE_RSA_CHACHA20_POLY1305,
    Ciphers.CHACHA20_POLY1305_SHA256,
]

# List of ciphers containing a small variety of symmetric ciphers
# with a small variety of auth methods.
MINIMAL_TEST_CIPHERS = [
    Ciphers.DHE_RSA_AES128_SHA,
    Ciphers.AES256_SHA,
    Ciphers.ECDHE_ECDSA_AES128_GCM_SHA256,
    Ciphers.ECDHE_RSA_AES256_GCM_SHA384,
    Ciphers.ECDHE_RSA_CHACHA20_POLY1305,
    Ciphers.CHACHA20_POLY1305_SHA256,
    Ciphers.AES256_GCM_SHA384,
]

# List of TLS13 Ciphers
TLS13_CIPHERS = [
    Ciphers.CHACHA20_POLY1305_SHA256,
    Ciphers.AES128_GCM_SHA256,
    Ciphers.AES256_GCM_SHA384,
]

S2N_TEST_POLICIES = [
    Ciphers.SECURITY_POLICY_DEFAULT,
    Ciphers.SECURITY_POLICY_DEFAULT_TLS12,
    Ciphers.SECURITY_POLICY_DEFAULT_TLS13,
]


# List of ports available to tests.
available_ports = AvailablePorts()


# Server certificates used to test matching domain names client with server_name
# ( cert_path, private_key_path, domains[] )
SNI_CERTS = {
    "alligator": (
        TEST_SNI_CERT_DIRECTORY + "alligator_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "alligator_key.pem",
        ["www.alligator.com"],
    ),
    "second_alligator_rsa": (
        TEST_SNI_CERT_DIRECTORY + "second_alligator_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "second_alligator_rsa_key.pem",
        ["www.alligator.com"],
    ),
    "alligator_ecdsa": (
        TEST_SNI_CERT_DIRECTORY + "alligator_ecdsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "alligator_ecdsa_key.pem",
        ["www.alligator.com"],
    ),
    "beaver": (
        TEST_SNI_CERT_DIRECTORY + "beaver_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "beaver_key.pem",
        ["www.beaver.com"],
    ),
    "many_animals": (
        TEST_SNI_CERT_DIRECTORY + "many_animal_sans_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "many_animal_sans_rsa_key.pem",
        [
            "www.catfish.com",
            "www.dolphin.com",
            "www.elephant.com",
            "www.falcon.com",
            "www.gorilla.com",
            "www.horse.com",
            "www.impala.com",
            # "Simple hostname"
            "Jackal",
            "k.e.e.l.b.i.l.l.e.d.t.o.u.c.a.n",
            # SAN on this cert is actually "ladybug.ladybug"
            # Verify case insensitivity works as expected.
            "LADYBUG.LADYBUG",
            "com.penguin.macaroni",
        ],
    ),
    "narwhal_cn": (
        TEST_SNI_CERT_DIRECTORY + "narwhal_cn_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "narwhal_cn_key.pem",
        ["www.narwhal.com"],
    ),
    "octopus_cn_platypus_san": (
        TEST_SNI_CERT_DIRECTORY + "octopus_cn_platypus_san_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "octopus_cn_platypus_san_key.pem",
        ["www.platypus.com"],
    ),
    "quail_cn_rattlesnake_cn": (
        TEST_SNI_CERT_DIRECTORY + "quail_cn_rattlesnake_cn_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "quail_cn_rattlesnake_cn_key.pem",
        ["www.quail.com", "www.rattlesnake.com"],
    ),
    "many_animals_mixed_case": (
        TEST_SNI_CERT_DIRECTORY + "many_animal_sans_mixed_case_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "many_animal_sans_mixed_case_rsa_key.pem",
        [
            "alligator.com",
            "beaver.com",
            "catFish.com",
            "WWW.dolphin.COM",
            "www.ELEPHANT.com",
            "www.Falcon.Com",
            "WWW.gorilla.COM",
            "www.horse.com",
            "WWW.IMPALA.COM",
            "WwW.jAcKaL.cOm",
        ],
    ),
    "embedded_wildcard": (
        TEST_SNI_CERT_DIRECTORY + "embedded_wildcard_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "embedded_wildcard_rsa_key.pem",
        ["www.labelstart*labelend.com"],
    ),
    "non_empty_label_wildcard": (
        TEST_SNI_CERT_DIRECTORY + "non_empty_label_wildcard_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "non_empty_label_wildcard_rsa_key.pem",
        ["WILD*.middle.end"],
    ),
    "trailing_wildcard": (
        TEST_SNI_CERT_DIRECTORY + "trailing_wildcard_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "trailing_wildcard_rsa_key.pem",
        ["the.prefix.*"],
    ),
    "wildcard_insect": (
        TEST_SNI_CERT_DIRECTORY + "wildcard_insect_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "wildcard_insect_rsa_key.pem",
        [
            "ant.insect.hexapod",
            "BEE.insect.hexapod",
            "wasp.INSECT.hexapod",
            "butterfly.insect.hexapod",
        ],
    ),
    "termite": (
        TEST_SNI_CERT_DIRECTORY + "termite_rsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "termite_rsa_key.pem",
        ["termite.insect.hexapod"],
    ),
    "underwing": (
        TEST_SNI_CERT_DIRECTORY + "underwing_ecdsa_cert.pem",
        TEST_SNI_CERT_DIRECTORY + "underwing_ecdsa_key.pem",
        ["underwing.insect.hexapod"],
    ),
}


# Test cases for certificate selection.
# Test inputs: server certificates to load into s2nd, client SNI and capabilities, outputs are selected server cert
# and negotiated cipher.
MultiCertTest = collections.namedtuple(
    "MultiCertTest",
    "description server_certs client_sni client_ciphers expected_cert expect_matching_hostname",
)
MULTI_CERT_TEST_CASES = [
    MultiCertTest(
        description="Test basic SNI match for default cert.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test basic SNI matches for non-default cert.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni="www.beaver.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["beaver"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test default cert is selected when there are no SNI matches.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni="not.a.match",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test default cert is selected when no SNI is sent.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni=None,
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test ECDSA cert is selected with matching domain and client only supports ECDSA.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.ECDHE_ECDSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator_ecdsa"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test ECDSA cert selected when: domain matches for both ECDSA+RSA, client supports ECDSA+RSA "
        " ciphers, ECDSA is higher priority on server side.",
        server_certs=[
            SNI_CERTS["alligator"],
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator_ecdsa"],
        ],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA, Ciphers.ECDHE_ECDSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator_ecdsa"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test domain match is highest priority. Domain matching ECDSA certificate should be selected"
        " even if domain mismatched RSA certificate is available and RSA cipher is higher priority.",
        server_certs=[SNI_CERTS["beaver"], SNI_CERTS["alligator_ecdsa"]],
        client_sni="www.alligator.com",
        client_ciphers=[
            Ciphers.ECDHE_RSA_AES128_SHA256,
            Ciphers.ECDHE_ECDSA_AES128_SHA256,
        ],
        expected_cert=SNI_CERTS["alligator_ecdsa"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test certificate with single SAN entry matching is selected before mismatched multi SAN cert",
        server_certs=[SNI_CERTS["many_animals"], SNI_CERTS["alligator"]],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=True,
    ),
    # many_animals was the first cert added
    MultiCertTest(
        description="Test default cert with multiple sans and no SNI sent.",
        server_certs=[SNI_CERTS["many_animals"], SNI_CERTS["alligator"]],
        client_sni=None,
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["many_animals"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test certificate match with CN",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["narwhal_cn"]],
        client_sni="www.narwhal.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["narwhal_cn"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test SAN+CN cert can match using SAN.",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["octopus_cn_platypus_san"]],
        client_sni="www.platypus.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["octopus_cn_platypus_san"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Test that CN is not considered for matching if the certificate contains SANs.",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["octopus_cn_platypus_san"]],
        client_sni="www.octopus.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test certificate with multiple CNs can match.",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["quail_cn_rattlesnake_cn"]],
        client_sni="www.rattlesnake.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["quail_cn_rattlesnake_cn"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test cert with embedded wildcard is not treated as a wildcard.",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["embedded_wildcard"]],
        client_sni="www.labelstartWILDCARDlabelend.com",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test non empty left label wildcard cert is not treated as a wildcard."
        " s2n only supports wildcards with a single * as the left label",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["non_empty_label_wildcard"]],
        client_sni="WILDCARD.middle.end",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Test cert with trailing * is not treated as wildcard.",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["trailing_wildcard"]],
        client_sni="the.prefix.WILDCARD",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=False,
    ),
    MultiCertTest(
        description="Certificate with exact sni match(termite.insect.hexapod) is preferred over wildcard"
        " *.insect.hexapod",
        server_certs=[
            SNI_CERTS["wildcard_insect"],
            SNI_CERTS["alligator"],
            SNI_CERTS["termite"],
        ],
        client_sni="termite.insect.hexapod",
        client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
        expected_cert=SNI_CERTS["termite"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="ECDSA Certificate with exact sni match(underwing.insect.hexapod) is preferred over RSA wildcard"
        " *.insect.hexapod when RSA ciphers are higher priority than ECDSA in server preferences.",
        server_certs=[
            SNI_CERTS["wildcard_insect"],
            SNI_CERTS["alligator"],
            SNI_CERTS["underwing"],
        ],
        client_sni="underwing.insect.hexapod",
        client_ciphers=[
            Ciphers.ECDHE_RSA_AES128_GCM_SHA256,
            Ciphers.ECDHE_ECDSA_AES128_GCM_SHA256,
        ],
        expected_cert=SNI_CERTS["underwing"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Firstly loaded matching certificate should be selected among certificates with the same domain names",
        server_certs=[SNI_CERTS["alligator"], SNI_CERTS["second_alligator_rsa"]],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.AES128_GCM_SHA256],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=True,
    ),
    MultiCertTest(
        description="Firstly loaded matching certificate should be selected among matching+non-matching certificates",
        server_certs=[
            SNI_CERTS["beaver"],
            SNI_CERTS["alligator"],
            SNI_CERTS["second_alligator_rsa"],
        ],
        client_sni="www.alligator.com",
        client_ciphers=[Ciphers.AES128_GCM_SHA256],
        expected_cert=SNI_CERTS["alligator"],
        expect_matching_hostname=True,
    ),
]
# Positive test for wildcard matches
MULTI_CERT_TEST_CASES.extend(
    [
        MultiCertTest(
            description="Test wildcard *.insect.hexapod matches subdomain "
            + specific_insect_domain,
            server_certs=[SNI_CERTS["alligator"], SNI_CERTS["wildcard_insect"]],
            client_sni=specific_insect_domain,
            client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
            expected_cert=SNI_CERTS["wildcard_insect"],
            expect_matching_hostname=True,
        )
        for specific_insect_domain in SNI_CERTS["wildcard_insect"][2]
    ]
)
# Positive test for basic SAN matches
MULTI_CERT_TEST_CASES.extend(
    [
        MultiCertTest(
            description="Match SAN " + many_animal_domain + " in many_animals cert",
            server_certs=[SNI_CERTS["alligator"], SNI_CERTS["many_animals"]],
            client_sni=many_animal_domain,
            client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
            expected_cert=SNI_CERTS["many_animals"],
            expect_matching_hostname=True,
        )
        for many_animal_domain in SNI_CERTS["many_animals"][2]
    ]
)
# Positive test for mixed cased SAN matches
MULTI_CERT_TEST_CASES.extend(
    [
        MultiCertTest(
            description="Match SAN "
            + many_animal_domain
            + " in many_animals_mixed_case cert",
            server_certs=[SNI_CERTS["alligator"], SNI_CERTS["many_animals_mixed_case"]],
            client_sni=many_animal_domain,
            client_ciphers=[Ciphers.ECDHE_RSA_AES128_SHA],
            expected_cert=SNI_CERTS["many_animals_mixed_case"],
            expect_matching_hostname=True,
        )
        for many_animal_domain in SNI_CERTS["many_animals_mixed_case"][2]
    ]
)
